-- $Source: /home/dsd/d/aa/monoBANK/AA/ada.direct_io.spc,v $
-- $Revision: 1.4 $ -- $Date: 96/03/15 10:29:13 $ -- $Author: akira $
--        Copyright (C) 1995 by Intermetrics, Inc.   

pragma Suppress( Elaboration_Check );

with System.RTS.TGT.Basic_IO;
with Ada.IO_Exceptions;

pragma ada_child;
generic
    type Element_Type is private;

package Ada.Direct_IO is

    --| Overview
    --|
    --| This package implements Direct_IO, as specified in the Ada
    --| Language Reference Manual (LRM), section 14.2.4.

    type File_Type is limited private;

    type File_Mode is (In_File, Out_File, InOut_File);

    -- The definition of type count may need to be changed
    -- for systems which do not support "large" files

    type Count is range 0 .. integer'last;
    subtype Positive_Count is Count range 1 .. Count'last;

    -- File management

    procedure Create(file : in out file_type;
                     mode : in file_mode := inout_file;
		     name : in string := "";
		     form : in string := "");

    --| Effect
    --|
    --| Establishes a new external file with the given name and form,
    --| and associates this external file with the given file.
    --| The given file is left open. Raises STATUS_ERROR if the given
    --| file is already open. Raises NAME_ERROR if the string given
    --| as NAME does not allow the identification of an external
    --| file.

    procedure Open(file : in out file_type;
                   mode : in file_mode;
		   name : in string;
		   form : in string := "");

    --| Effect
    --|
    --| Associates the given file with an existing external file having
    --| the given name and form, and sets the current mode of the file to
    --| the given mode. The given file is left open.
    --|
    --| The exception STATUS_ERROR is raised if the given file is
    --| already open. The exception NAME_ERROR is raised if the
    --| string given as NAME does not allow the identification of
    --| an external file.


    procedure Close(file : in out file_type);

    --| Effect
    --|
    --| Severs the association between the given file and it's
    --| associated external file. The given file is left closed.
    --|
    --| The exception STATUS_ERROR is raised if the given file is not
    --| open.

    procedure Delete(file : in out file_type);

    --| Effect
    --|
    --| Deletes the external file associated with the given file.
    --| The given file is closed, and the external file ceases to
    --| exist.
    --|
    --| The exception STATUS_ERROR is raised if the given file is not
    --| open. The exception USE_ERROR is raised if deletion of the
    --| external file is not provided by the environment.

    procedure Reset(file : in out file_type; mode : in file_mode);
    procedure Reset(file : in out file_type);

    --| Effect
    --|
    --| Resets the given file so that reading from or writing to its
    --| elements can be restarted from the beginning of the file; for
    --| a direct access file this implies that the current index is
    --| set to one. If a mode parameter is supplied, the current mode
    --| of the given file is set to the given mode.
    --|
    --| The exception STATUS_ERROR is raised if the file is not open.
    --| The exception USE_ERROR is raised if the environment does not
    --| support resetting for the external file and, also, if the
    --| environment does not support resetting to the specified mode
    --| for the external file.


    function Mode(file : in file_type) return file_mode;

    --| Effect
    --|
    --| Returns the current mode of the given file.
    --|
    --| The exception STATUS_ERROR is raised if the given file is
    --| not open.

    function Name(file : in file_type) return string;

    --| Effect
    --|
    --| Returns a string which uniquely identifies the external file
    --| currently associated with the given file (and may thus be used
    --| in an OPEN operation).
    --|
    --| The exception STATUS_ERROR is raised if the given file is not
    --| open.

    function Form(file : in file_type) return string;

    --| Effect
    --|
    --| Returns the form string for the external file currently
    --| associated with the given file.
    --|
    --| The exception STATUS_ERROR is raised if the given file is
    --| not open.


    function Is_Open(File : in File_Type) return boolean;

    --| Effect
    --|
    --| Returns TRUE if the file is open, otherwise returns false.


    -- Input and output operations

    procedure Read(file : in file_type; item : out element_type;
                   from : positive_count);
    procedure Read(file : in file_type; item : out element_type);

    --| Effect
    --|
    --| Operates on a file of mode IN_FILE or INOUT_FILE. Returns,
    --| in the parameter ITEM, the value of the element whose
    --| position in the given file is specified by the current
    --| index of the file; finally, increases the current index
    --| by one.
    --|
    --| The exception MODE_ERROR is raised if the mode of the given
    --| file is OUT_FILE. The exception END_ERROR is raised if the
    --| index to be used exceeds the size of the external file. The
    --| exception DATA_ERROR is raised if the element read cannot
    --| be interpreted as a value of the type ELEMENT_TYPE.


    procedure Write(file : in file_type; item : in element_type;
                    to : positive_count);
    procedure Write(file : in file_type; item : in element_type);

    --| Effect
    --|
    --| Operates on a file of mode OUT_FILE or INOUT_FILE. In the case
    --| of the first form, sets the index of the given file to the
    --| index value given by the parameter TO. Then (for both forms)
    --| gives the value of the parameter ITEM to the element whose
    --| position in the given file is specified by the current index
    --| of the file; finally increases the current index by one.
    --|
    --| The exception MODE_ERROR is raised if the mode of the given file
    --| is IN_FILE. The exception USE_ERROR is raised if the capacity
    --| of the external file is exceeded.

    procedure Set_Index(file : in file_type; to : in positive_count);

    --| Effect
    --|
    --| Operates on a file of any mode. Sets the current index of the
    --| given file to the given index value (which may exceed the
    --| current size of the file).

    function Index(file : in file_type) return positive_count;

    --| Effect
    --|
    --| Operates on a file of any mode. Returns the current index of
    --| the give file.

    function Size(file : in file_type) return count;

    --| Effect
    --|
    --| Operates on a file of any mode. Returns the current size of
    --| the external file that is associated with the given file.

    function End_Of_File(file : in file_type) return boolean;

    --| Effect
    --|
    --| Operates on a file of mode IN_FILE or INOUT_FILE. Returns TRUE
    --| if the current index exceeds the size of the external file;
    --| otherwise returns FALSE.
    --|
    --| The exception MODE_ERROR is raised if the mode of the given file
    --| is OUT_FILE.

    -- Exceptions

    Status_Error : exception renames Ada.IO_Exceptions.Status_Error;
    Mode_Error   : exception renames Ada.IO_Exceptions.Mode_Error;
    Name_Error   : exception renames Ada.IO_Exceptions.Name_Error;
    Use_Error    : exception renames Ada.IO_Exceptions.Use_Error;
    Device_Error : exception renames Ada.IO_Exceptions.Device_Error;
    End_Error    : exception renames Ada.IO_Exceptions.End_Error;
    Data_Error   : exception renames Ada.IO_Exceptions.Data_Error;

private

    type File_Record;

    type File_Type is access File_Record;

    type File_Record is
        record
            Mode   : File_Mode;
            Handle : System.RTS.TGT.Basic_IO.File_Type;
        end record;

end Ada.Direct_IO;

with Ada.Direct_IO;
generic package Direct_IO renames Ada.Direct_IO;
