
-- ***************************************************************************
--
--  COPYRIGHT NOTICE:
--    This software is protected as an unpublished work under the Copyright
--    Act of 1976. Created 1986 (C). All rights reserved. Alsys.
--    "Alsys" is a trademark of Alsys, S.A.
--
--  PACKAGE SPECIFICATION: UNSIGNED
--  -------------------------------
--
-- ***************************************************************************

with UNCHECKED_CONVERSION;

package UNSIGNED is

-- ***************************************************************************
--
--  Outline:
--  --------
--
--     This package provides types that implement unsigned bytes and words
--     and related operations. At this release of the compiler (version 3.1)
--     representation clauses are implemented. Thus, the user could specify 
--     their own unsigned types. 
--
--  Contents:
--  ---------
--
--     (1)  The types BYTE and WORD
--     (2)  Conversions
--     (3)  Unsigned arithmetic functions
--          (Unsigned "modulus" arithmetic)
--
-- ***************************************************************************



--	    ********************************
--	    * (1)  The types BYTE and WORD *
--	    ********************************

   subtype INTEGER_8 is SHORT_SHORT_INTEGER;
   subtype INTEGER_16 is SHORT_INTEGER;
   subtype INTEGER_32 is INTEGER;

   type BYTE is mod 2**8;
   type WORD is mod 2**16;

   -- When used as the type for a component of a record, or as the element
   -- type of an array, a BYTE is allocated only 8 bits (instead of 16 bits
   -- it would normally get as a subtype of INTEGER) and a word is allocated
   -- only 16 bits (instead of the 32 bits it would normally get as a subtype
   -- of LONG_INTEGER).  When allocated as a scalar object in a frame, a word
   -- is allocated only 16 bits.

   -- Any type derived from BYTE or WORD, or any subtype of BYTE or WORD,
   -- will have the same property.

   -- A string cannot be directly indexed by an object of type BYTE or WORD,
   -- since they are new types and a string can only be indexed by a NATURAL.
   -- However, a simple conversion to NATURAL does the trick.




   --	    *****************************
   --	    * (2) Conversion Operations *
   --	    *****************************


   function TO_SHORT_INTEGER is new UNCHECKED_CONVERSION (BYTE, INTEGER_8);

   -- TO_SHORT_INTEGER takes a byte and returns an integer, assuming a
   -- 8 bit twos complement view of the byte value (e.g., if the argument
   -- is 255 the returned result is -1).


   function TO_BYTE is new UNCHECKED_CONVERSION (INTEGER_8, BYTE);

   -- TO_BYTE performs the reverse conversion.  The argument is a short
   -- integer in the range -128 .. +127.  The result is the corresponding
   -- unsigned byte value in twos complement form (e.g., if the input is
   -- -127 then the output is 129.


   function TO_CHARACTER is new UNCHECKED_CONVERSION (CHARACTER, BYTE);
   -- TO_CHARACTER takes a character and returns a byte.  In contrast to 
   -- Ada83 this is a safe operation.

   function TO_BYTE is new UNCHECKED_CONVERSION (CHARACTER, BYTE);

   -- TO_BYTE performs the reverse conversion.  It is essentially
   -- equivalent to CHARACTER'POS (C)


   function TO_INTEGER is new UNCHECKED_CONVERSION (WORD, INTEGER_16);

   -- TO_INTEGER takes a word and returns an integer, assuming a 16 bit
   -- twos complement view of the word value (e.g. if the argument is 65535,
   -- then the returned result is -1).


   function TO_WORD is new UNCHECKED_CONVERSION (INTEGER_16, WORD);

   -- TO_WORD performs the reverse conversion. The argument is an integer
   -- in the range -32768 to +32767. The result is the corresponding unsigned
   -- word value in twos complement form (e.g. if the input is -32767, then
   -- the result will be 32769).


   function LSB (W: WORD) return BYTE;

   -- LSB returns the least significant byte of its word argument as a byte


   function MSB (W: WORD) return BYTE;

   -- MSB returns the most significant byte of its word argument as a byte


   function MAKE_WORD (MSB: BYTE; LSB: BYTE) return WORD;

   -- MAKE_WORD returns a word given two bytes. The relationship between this
   -- function and the others is that MAKE_WORD(MSB(W),LSB(W)) = W always holds.


   -- Note: the following Ada type conversions can also be used:

   --    BYTE (WORD)	Converts WORD value in range 0..255 to BYTE
   --    WORD (BYTE)	Converts BYTE value to WORD
   --    BYTE (INTEGER)	Converts INTEGER value in range 0..255 to BYTE
   --    INTEGER (BYTE)	Converts BYTE value to INTEGER
   --    WORD (INTEGER)	Converts INTEGER value in range 0..32767 to WORD
   --    INTEGER (WORD)	Converts WORD value in range 0..32767 to INTEGER


   --	    ************************
   --	    * (3) Shift operations *
   --	    ************************

   -- Note; the type of RIGHT has been changed from INTEGER to NATURAL

   function SHL_8 (LEFT : BYTE; RIGHT : NATURAL) return BYTE;
   pragma Import (Intrinsic, SHL_8, Link_Name => "C_SHL_8" );
   function SHL (LEFT : BYTE; RIGHT : NATURAL) return BYTE
      renames SHL_8;

   function SHR_8 (LEFT : BYTE; RIGHT : NATURAL) return BYTE;
   pragma Import (Intrinsic, SHR_8, Link_Name => "C_SHR_8" );
   function SHR (LEFT : BYTE; RIGHT : NATURAL) return BYTE
      renames SHR_8;

   function SHL_16 (LEFT : WORD; RIGHT : NATURAL) return WORD;
   pragma Import (Intrinsic, SHL_16, Link_Name => "C_SHL_16" );
   function SHL (LEFT : WORD; RIGHT : NATURAL) return WORD
      renames SHL_16;

   function SHR_16 (LEFT : WORD; RIGHT : NATURAL) return WORD;
   pragma Import (Intrinsic, SHR_16, Link_Name => "C_SHR_16" );
   function SHR (LEFT : WORD; RIGHT : NATURAL) return WORD
      renames SHR_16;

   --	    *******************************************
   --	    * (4) Modulus Arithmetic on BYTE and WORD *
   --	    *******************************************

   function "+" (OPR: INTEGER_16) return BYTE;
   function "+" (OPR: INTEGER_32) return WORD;

private

   pragma INLINE ("+", LSB, MSB, MAKE_WORD);
   
end UNSIGNED;
