-- $Source: /home/dsd/d/aa/monoBANK/AA/ada.direct_io.bdy,v $ 
-- $Revision: 1.5 $ $Date: 95/10/12 16:50:50 $ $Author: akira $ 
--        Copyright (C) 1995 by Intermetrics, Inc.   

with Ada.Unchecked_Deallocation;

pragma Elaborate(System.RTS.TGT.Basic_IO);

pragma ada_child;
package body Ada.Direct_IO is
    
    pragma Suppress( Elaboration_Check );

    package BIO renames System.RTS.TGT.Basic_IO;

    Element_Size : constant Natural :=
      (Element_Type'SIZE + (BIO.Bits_Per_IO_Element - 1)) /
        BIO.Bits_Per_IO_Element;
      -- I/O element type size in storage units

    Internal_Error : exception;

    DIO_Mode_To_BIO_Mode : constant array(File_Mode) of
      BIO.File_Mode := (In_File    => BIO.In_File,
                        Out_File   => BIO.Out_File,
                        InOut_File => BIO.InOut_File);

    BIO_Mode_To_DIO_Mode : constant array(BIO.File_Mode range
      BIO.In_File .. BIO.InOut_File) of File_Mode :=
       (BIO.In_File    => In_File,
        BIO.Out_File   => Out_File,
        BIO.InOut_File => InOut_File);

    procedure File_Free is new
      Unchecked_Deallocation(File_Record, File_Type);


    procedure Create (File : in out File_Type;
                      Mode : in File_Mode := InOut_File;
                      Name : in String := "";
                      Form : in String := "") is
        Handle : BIO.File_Type;
    begin
        if File /= null then
            raise Status_Error;
        end if;
        BIO.Create(Handle,
                   DIO_Mode_To_BIO_Mode(Mode), Name, Form,
                   BIO.Direct_IO,
                   Element_Size);
        File := new File_Record'(Mode => Mode, Handle => Handle);
    end Create;

    procedure Open (File : in out File_Type;
                    Mode : in File_Mode;
                    Name : in String;
                    Form : in String := "") is
        Handle : BIO.File_Type;
    begin
        if File /= null then
            raise Status_Error;
        end if;
        BIO.Open(Handle,
                 DIO_Mode_To_BIO_Mode(Mode), Name, Form,
                 BIO.Direct_IO,
                 Element_Size);
        File := new File_Record'(Mode => Mode, Handle => Handle);
    end Open;

    procedure Close(File : in out File_Type) is
    begin
        if File = null then
            raise Status_Error;
        end if;
        BIO.Close(File.Handle);
        File_Free(File);
        File := null;
    end Close;

    procedure Delete(File : in out File_Type) is
    begin
        if File = null then
            raise Status_Error;
        end if;
        BIO.Delete(File.Handle);
        File_Free(File);
        File := null;
    end Delete;

    procedure Reset(File : in out File_Type;
                    Mode : in File_Mode) is
    begin
        if File = null then
            raise Status_Error;
        end if;
        BIO.Switch_Mode(File => File.Handle,
                        Old_Mode => DIO_Mode_To_BIO_Mode(File.Mode),
                        New_Mode => DIO_Mode_To_BIO_Mode(Mode));
        -- Update file record to reflect new mode
        if File.Mode = Mode then
            -- Mode wasn't changed, so we need to reposition file pointer
            -- explicitly
            BIO.Reset(File.Handle);
        else
            -- Mode was changed, so Switch_Mode repositioned
            -- file pointer for us.
            File.Mode := Mode;
        end if;
    end Reset;

    procedure Reset(File : in out File_Type) is
    begin
        if File = null then
            raise Status_Error;
        end if;
        BIO.Reset(File.Handle);
    end Reset;

    function Mode(File : in File_Type) return File_Mode is
    begin
        if File = null then
            raise Status_Error;
        end if;
        return File.Mode;
    end Mode;

    function Name(File : in File_Type) return String is
    begin
        if File = null then
            raise Status_Error;
        end if;
        return BIO.Name(File.Handle);
    end Name;

    function Form(File : in File_Type) return String is
    begin
        if File = null then
            raise Status_Error;
        end if;
        return BIO.Form(File.Handle);
    end Form;

    function Is_Open (File : in File_Type) return Boolean is
    begin
        return File /= null;
    end Is_Open;


    procedure Read(File : in File_Type;
                   Item : out Element_Type) is
        Amount_Read : Natural;
    begin
        if File = null then
            raise Status_Error;
        end if;
        if File.Mode = Out_File then
            raise Mode_Error;
        end if;
        Amount_Read := BIO.Read(File.Handle,
                                Item'Address,
                                Element_Size);
        if Amount_Read = 0 then
            raise End_Error;
        end if;
    end Read;

    procedure Read(File : in File_Type;
                   Item : out Element_Type;
                   From : Positive_Count) is
    begin
        Set_Index(File, From);
        Read(File, Item);
    end Read;

    procedure Write(File : in File_Type;
                    Item : in Element_Type) is
        Amount_Written : Natural;
    begin
        if File = null then
            raise Status_Error;
        end if;
        if File.Mode = In_File then
            raise Mode_Error;
        end if;
        Amount_Written := BIO.Write(File.Handle,
                                    Item'Address,
				    Element_Size);
        if Amount_Written /= Element_Size then
            raise Internal_Error;
        end if;
    end Write;

    procedure Write(File : in File_Type;
                    Item : in Element_Type;
                    To   : Positive_Count) is
    begin
        Set_Index(File, To);
        Write(File, Item);
    end Write;

    procedure Set_Index(File : in File_Type;
                        To   : in Positive_Count) is
    begin
        if File = null then
            raise Status_Error;
        end if;
        BIO.Set_Index(File.Handle,
                      Integer(To));
    end Set_Index;

    function Index(File : in File_Type) return Positive_Count is
    begin
        if File = null then
            raise Status_Error;
        end if;
        return Positive_Count(BIO.Index(File.Handle));
    end Index;

    function Size ( File : in File_Type) return Count is
    begin
        if File = null then
            raise Status_Error;
        end if;
        return Count(BIO.Size(File.Handle));
    end Size;

    function End_Of_File(File : in File_Type) return Boolean is
    begin
        if File = null then
            raise Status_Error;
        end if;
        if File.Mode = Out_File then
            raise Mode_Error;
        end if;
        return Index(File) > Size(File);
    end End_Of_File;


end Ada.Direct_IO;
