-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-- This unpublished work is protected both as a proprietary work and under
-- the Universal Copyright Convention and the US Copyright Act of 1976.  Its
-- distribution and access are limited only to authorized persons.  Copyright
-- (C) Alsys.  Created 1990, initially licenced 1990.  All rights reserved.
-- Unauthorized use (including use to prepare other works), disclosure,
-- reproduction or distribution may violate national criminal law.
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

------------------------------------------------------------------------------
-- PACKAGE BODY System.RTS.TGT.Exceptions.TDEP	 (TEMPLATE FILE)
-- -------------------------------------------
-- This module defines the target dependent data structures and procedures  --
-- for the common exception manager.
------------------------------------------------------------------------------

with System.Rts.Error_IO;
with System.RTS.TGT.Exceptions.PROP;
with System.RTS.TGT.Exceptions.TYPES;
with System.RTS.Exception_Types;
with SYSTEM;
with UNCHECKED_CONVERSION;
with SYSTEM.RTS.TCB;

package body System.RTS.TGT.Exceptions.TDEP is

   pragma SUPPRESS (ELABORATION_CHECK);
   pragma SUPPRESS (STORAGE_CHECK);

   --	    *********************
   --	    * TABLE OF CONTENTS *
   --	    *********************

   --	      1. PROPAGATE_CONTEXT
   --	      2. EXCEPTION_LOCATION access
   --	      3. GET_EXCEPTION_FRAME_PTR
   --	      4. EXITING_OUTER_FRAME
   --	      5. IS_ENVIRONMENT_TASK
   --	      6. DUMP_PROCEDURE_INFO_RECORD
   --	      7. DUMP_EXCEPTION_CONTEXT
   --	      Revision History

   package ERROR_IO renames System.Rts.Error_IO;
   package PROP renames System.RTS.TGT.Exceptions.PROP;
   package ETYPES renames System.RTS.Exception_Types;
   
   type REF_DATA_ADDRESS is access all DATA_ADDRESS;
   function FETCH is new
      UNCHECKED_CONVERSION (DATA_ADDRESS, REF_DATA_ADDRESS);
      

   --	    **************************
   --	    * 5. IS_ENVIRONMENT_TASK *
   --	    **************************


   function IS_ENVIRONMENT_TASK (CONTEXT : REF_EXCEPTION_CONTEXT)
							   return BOOLEAN is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- IS_ENVIRONMENT_TASK is used to determine whether the task for which
   -- EXITING_OUTER_FRAME is TRUE is an Ada task or the environment task.
   -- The Exception Manager invokes program shutdown actions if
   -- IS_ENVIRONMENT_TASK returns TRUE.
   ---------------------------------------------------------------------------
      use SYSTEM.RTS.TCB;
   begin
      return CURR_TCB = ENVIRONMENT_TCB;
   end;


   --	    **************************
   --	    * 4. EXITING_OUTER_FRAME *
   --	    **************************


   function EXITING_OUTER_FRAME (CONTEXT : REF_EXCEPTION_CONTEXT)
							   return BOOLEAN is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- EXITING_OUTER_FRAME is used to determine when propagation past the
   -- outermost frame of the Ada task or environment task has taken place,
   -- ie. it indicates task and program death. If this has occurred, a stack
   -- traceback is generated if the appropriate history option is set.
   ---------------------------------------------------------------------------
      ADA_ENV_TASK_SP : DATA_ADDRESS;
      pragma Import (C, ADA_ENV_TASK_SP, "ada_env_task_sp");
   begin
      -- For the Adamagic runtime, a task body always has a handler.
      -- So the only test that is needed is for exiting the environment
      -- task. 
      return IS_ENVIRONMENT_TASK (CONTEXT) and then CONTEXT.EBP >= ADA_ENV_TASK_SP;
   end;


   --	    ************************
   --	    * 1. PROPAGATE_CONTEXT *
   --	    ************************

     

   procedure PROPAGATE_CONTEXT (INFO	   : PROCEDURE_INFORMATION_RECORD;
				CONTEXT    : REF_EXCEPTION_CONTEXT;
				NEW_MARKER : in out DATA_ADDRESS) is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- PROPAGATE_CONTEXT uses the supplied PROCEDURE_INFORMATION_RECORD to
   -- propagate the context to the calling procedure, modifying CONTEXT.
   -- In addition, NEW_MARKER is set to the appropriate value of the new
   -- clean up marker (i.e. to a value which will cause cleanup of all
   -- data in the called procedure whose frame is destroyed by the call).
   -- The cleanup marker may be left unchanged from its input value if
   -- there is no additional data to cleanup as a result of the propagation.
   ---------------------------------------------------------------------------
      STARTING_EBP         : DATA_ADDRESS := CONTEXT.EBP;
      STARTING_ESP         : DATA_ADDRESS := CONTEXT.ESP;
      PREVIOUS_EBP         : DATA_ADDRESS := FETCH (CONTEXT.EBP).all;
      PREVIOUS_TCB         : DATA_ADDRESS;
      SAVED_DISPLAY_OFFSET : SYSTEM.OFFSET;
      
      function REMOVE_EXCEPTION_IN_PROPAGATE (
         ESTABLISHING_EBP : DATA_ADDRESS)
         return DATA_ADDRESS;
      pragma Import (Assembler, REMOVE_EXCEPTION_IN_PROPAGATE,
                     "_rts_remove_exception_in_propagate");
      -- Removes the exception registration record established by the
      -- subprogram at the given frame. Returns the new current 
      -- tcb pointer or 0 if the tcb pointer did not change.

   begin
      if STARTING_EBP >= PREVIOUS_EBP then
         -- Stack should propagate down
         CONTEXT.EIP := NULL_CODE_ADDRESS;  -- force end of propagation
      else

         if INFO.SAVED_DISPLAY_LEVEL > 0 then
            -- Restore display
            SAVED_DISPLAY_OFFSET := 4;
            if INFO.EXPORT then
               -- add size of exception record/soc/tcb save area
               SAVED_DISPLAY_OFFSET := 28+4;
            end if;
            CONTEXT.DISPLAY (INFO.SAVED_DISPLAY_LEVEL-1) := 
               FETCH (CONTEXT.EBP - SAVED_DISPLAY_OFFSET).all;
         end if;
         CONTEXT.EIP := CODE_ADDRESS (FETCH (CONTEXT.EBP+4).all) - 1;
         -- Subtract 1 to put EIP on the previous instruction (the CALL)
         CONTEXT.ESP := STARTING_EBP+8+
                        SYSTEM.OFFSET (INFO.PARAMETER_SIZE);
         -- skip past old EBP, old EIP, and parameters on stack
         CONTEXT.EBP := PREVIOUS_EBP;

         if INFO.EXPORT then
            -- We must undo the special actions taken in the subprogram prolog.
               
            -- Remove the exception registration record and restore tcb if necessary.
            PREVIOUS_TCB := REMOVE_EXCEPTION_IN_PROPAGATE (STARTING_EBP);
            if PREVIOUS_TCB /= System.Rts.Types.NULL_DATA_ADDRESS then
               -- We are going out of a DLL-exported subprogram.
               CONTEXT.TCB_POINTER := PREVIOUS_TCB;
               -- If we are propagating into an Ada DLL, then we must refetch a 
               -- copy of the display into CONTEXT, so it can be updated correctly.
               declare 
                  function TO_REF_TCB is 
                     new UNCHECKED_CONVERSION (DATA_ADDRESS, SYSTEM.RTS.TCB.REF_TCB);
                  TCB_PTR : SYSTEM.RTS.TCB.REF_TCB := TO_REF_TCB (CONTEXT.TCB_POINTER);
                  use System.RTS.TGT.Exceptions.TYPES;
               begin
                  if PROP.FIND_ADA_CODE_TABLE (CONTEXT) /= null then
                     for I in CONTEXT.DISPLAY'RANGE loop
                        CONTEXT.DISPLAY (I) := TCB_PTR.TCB_DISPLAY (I);
                     end loop;
                  end if;
               end;
            end if;
         end if;
      end if;
      
   end PROPAGATE_CONTEXT;



   function PROPAGATE_CONTEXT (CONTEXT : REF_EXCEPTION_CONTEXT)
						    return BOOLEAN is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- This second version of PROPAGATE_CONTEXT is used to propagate the
   -- exception from a non-Ada frame. On entry, CONTEXT indicates that the
   -- PC is not in Ada code. Each target must consider whether it can
   -- propagate exceptions through non-Ada frames. Three options are
   -- supported:
   --	 (a) Propagation is not supported and the body aborts the program.
   --	     In this case, the call does not return.
   --	 (b) Propagation is not supported, but a stack traceback should be
   --	     generated first (if requested) followed by program abort.
   --	     In this case, the function should return FALSE.
   --	 (c) Propagation is supported. In this case, the function should
   --	     return TRUE and CONTEXT updated accordingly.
   ---------------------------------------------------------------------------
   begin
      -- See if we're at the outermost level of the task
      if EXITING_OUTER_FRAME (CONTEXT) then
         return FALSE;

      -- Do some sanity checks to see that the stack is formatted correctly.
      elsif CONTEXT.EBP >= FETCH (CONTEXT.EBP).all then
         -- Stack should propagate down; can we do any other checks??
         return FALSE;
      else
         CONTEXT.EIP := CODE_ADDRESS (FETCH (CONTEXT.EBP+4).all) - 1;
         -- Subtract 1 to put EIP on the previous instruction (the CALL)
         CONTEXT.ESP := CONTEXT.EBP+8;
         CONTEXT.EBP := FETCH (CONTEXT.EBP).all;
         return TRUE;
      end if;
   end PROPAGATE_CONTEXT;


   --	    ********************************
   --	    * 2. EXCEPTION_LOCATION access *
   --	    ********************************

   function GET_EXCEPTION_LOCATION (CONTEXT : REF_EXCEPTION_CONTEXT)
						    return CODE_ADDRESS is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- GET_EXCEPTION_LOCATION obtains the current exception location from
   -- the exception context. It is typically just a field extraction.
   ---------------------------------------------------------------------------
   begin
      return CONTEXT.EIP;
   end;

   procedure UPDATE_CONTEXT (CONTEXT	     : REF_EXCEPTION_CONTEXT;
			     HANDLER_ADDRESS : CODE_ADDRESS) is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- UPDATE_CONTEXT updates the exception context with the supplied location
   -- (which is the address of the handler which we are about to go to).
   -- It is typically just a field update.
   ---------------------------------------------------------------------------
   begin
      CONTEXT.EIP := HANDLER_ADDRESS;
   end UPDATE_CONTEXT;


   --	    ******************************
   --	    * 3. GET_EXCEPTION_FRAME_PTR *
   --	    ******************************


   function GET_EXCEPTION_FRAME_PTR (INFO    : PROCEDURE_INFORMATION_RECORD;
				     CONTEXT : REF_EXCEPTION_CONTEXT)
						    return DATA_ADDRESS is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- GET_EXCEPTION_FRAME_PTR obtains the current frame pointer value
   -- from the exception context. It is typically just a field extraction.
   -- This is used in processing an HTE_SET_MARKER entry in the exception
   -- tables (i.e. the offset in such an entry is relative to the value
   -- returned by this function).
   ---------------------------------------------------------------------------
   begin
      return CONTEXT.EBP;
   end;



   function GET_ACTUAL_ECODE (ECODE : in EXCEPTION_CODE) 
                                                     return EXCEPTION_CODE is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- GET_ACTUAL_ECODE is used to covert what is potentially a pointer to
   -- an exception to an actual exception. ECODE can be a pointer to an
   -- exception if it names an imported exception.
   ---------------------------------------------------------------------------
      function TO_Exception_ID is 
         new UNCHECKED_CONVERSION (EXCEPTION_CODE, ETYPES.Exception_Id);

      type ECODE_ACCESS is access all EXCEPTION_CODE;
      function TO_ECODE_ACCESS is 
         new UNCHECKED_CONVERSION (EXCEPTION_CODE, ECODE_ACCESS);

      E  : Exception_Code := Ecode;
   begin
      while To_exception_id (E).Len >= 4096 loop
         E := TO_ECODE_ACCESS (E).all;
      end loop;
      return E;
   end GET_ACTUAL_ECODE;


   --	    *********************************
   --	    * 6. DUMP_PROCEDURE_INFO_RECORD *
   --	    *********************************


   procedure DUMP_PROCEDURE_INFO_RECORD (REC : PROCEDURE_INFORMATION_RECORD) is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- DUMP_PROCEDURE_INFO_RECORD is used in debug mode to dump the contents
   -- of a procedure information record. It should output a series of lines
   -- each preceded by 12 spaces, with one line feed at the end of the lines
   -- (i.e. NO extra blank line is generated before or after the output).
   ---------------------------------------------------------------------------
   begin
      ERROR_IO.PUT (RT_TYPES.LONGEST_INTEGER (REC.SAVED_DISPLAY_LEVEL));
   end;


   --	    *****************************
   --	    * 7. DUMP_EXCEPTION_CONTEXT *
   --	    *****************************

   procedure DUMP_EXCEPTION_CONTEXT (CONTEXT : REF_EXCEPTION_CONTEXT) is
   ---------------------------------------------------------------------------
   -- OUTLINE:
   -- -------
   -- DUMP_EXCEPTION_CONTEXT is used in debug mode to dump the contents
   -- of the exception context record. It should output a series of lines
   -- each preceded by 12 spaces, with one line feed at the end of the lines
   -- (i.e. NO extra blank line is generated before or after the output).
   ---------------------------------------------------------------------------
      I : DISPLAY_LEVEL;
   begin
      ERROR_IO.PUT ("            EIP = ");
      ERROR_IO.PUT (SYSTEM.ADDRESS (CONTEXT.EIP));
      ERROR_IO.NEW_LINE;
      ERROR_IO.PUT ("            EBP = ");
      ERROR_IO.PUT (SYSTEM.ADDRESS (CONTEXT.EBP));
      ERROR_IO.NEW_LINE;
      ERROR_IO.PUT ("            ESP = ");
      ERROR_IO.PUT (SYSTEM.ADDRESS (CONTEXT.ESP));
      ERROR_IO.NEW_LINE;
      I := DISPLAY_ARRAY'FIRST;
    DISPLAY_LOOP:
      while I < DISPLAY_ARRAY'LAST loop
         ERROR_IO.PUT ("            DISPLAY = ");
         for J in I .. I + 3 loop
             ERROR_IO.PUT (SYSTEM.ADDRESS (CONTEXT.DISPLAY (J)));
             if CONTEXT.DISPLAY (J) = NULL_DATA_ADDRESS then
                ERROR_IO.NEW_LINE;
                exit DISPLAY_LOOP;
             end if;
             ERROR_IO.PUT (" ");
         end loop;
         ERROR_IO.NEW_LINE;
         I := I + 4;
      end loop DISPLAY_LOOP;

   end DUMP_EXCEPTION_CONTEXT;


end System.RTS.TGT.Exceptions.TDEP;

--	    ********************
--	    * REVISION HISTORY *
--	    ********************

--  6.01  21 June 1991 (BJD)
--	  Initial version

--  6.02  3rd November 1991 (BJD)
--	  Prefix all CRTS library unit names with ALSYS_

--  6.03  11th May 1993 (BJD)
--	  Add second version of Propagate_Context for propagating through
--	  non-Ada frames.

--  6.04  3rd June 1993 (BJD)
--	  Added procedure Update_Context for use by EM_Prop.Exception_Handle
--	  to update a context with the handler address prior to outputting the
--	  corresponding entry in the traceback, and prior to transferring
--	  control to that handler context.
