-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-- This unpublished work is protected both as a proprietary work and under
-- the Universal Copyright Convention and the US Copyright Act of 1976.  Its
-- distribution and access are limited only to authorized persons.  Copyright
-- (C) Alsys.  Created 1990, initially licenced 1990.  All rights reserved.
-- Unauthorized use (including use to prepare other works), disclosure,
-- reproduction or distribution may violate national criminal law.
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-- ***************************************************************************
--
-- PACKAGE SPECIFICATION System.RTS.TGT.Exceptions.TDEP	    (TEMPLATE FILE)
-- ----------------------------------------------------	    ---------------
--
-- OUTLINE
-- -------
-- This module defines the target dependent data structures and procedures
-- for the exception manager.
--
-- ***************************************************************************

with System.Rts.Types; use System.Rts.Types;

package System.RTS.TGT.Exceptions.TDEP is

   pragma SUPPRESS (ELABORATION_CHECK);

   --	    *********************
   --	    * TABLE OF CONTENTS *
   --	    *********************

   --	      1. Type EXCEPTION_CONTEXT
   --	      2. Type PROCEDURE_INFORMATION_RECORD
   --	      3. Type HISTORY_TABLE
   --	      4. Subprograms
   --	      5. Private definitions of types
   --	      Revision History


   package RT_TYPES renames System.Rts.Types;


   --	    *****************************
   --	    * 1. TYPE EXCEPTION_CONTEXT *
   --	    *****************************


   -- EXCEPTION_CONTEXT is a private record containing the current exception
   -- context, including the location counter, and all saved over call
   -- registers, special registers (stack pointer etc.) and the display.

   type EXCEPTION_CONTEXT is private;
   type REF_EXCEPTION_CONTEXT is access EXCEPTION_CONTEXT;

   -- Note: although this type is private, assembly language glue routines
   -- are allowed to assume knowledge of the details of the contents and
   -- layout of this record, as described in the private part. The private
   -- declaration indicates and enforces the rule that the other parts of
   -- the Ada runtime, and in particular other parts of the exception manager,
   -- are NOT expected to mess with these structures (since they are target-
   -- dependent, and the rest of the exception manager is target-independent).


   --	    ****************************************
   --	    * 2. TYPE PROCEDURE_INFORMATION_RECORD *
   --	    ****************************************

   -- PROCEDURE_INFORMATION_RECORD is a private record containing sufficient
   -- information about a procedure to allow its calling context to be
   -- obtained when an exception is propagated.

   type PROCEDURE_INFORMATION_RECORD is private;

   -- Note: although this type is private, assembly language glue routines
   -- and the binder are allowed to assume knowledge of the details of the
   -- contents and layout of this record, as described in the private part. The
   -- The private declaration indicates and enforces the rule that the other
   -- parts of the Ada runtime, and in particular other parts of the exception
   -- manager, are NOT expected to mess with these structures (since they are
   -- target-dependent, and the rest of the exception manager is expected to
   -- be target-independent).

   -- Size of a PROCEDURE_INFORMATION_RECORD in bits. This is necessary since
   -- 'SIZE of a private type is not static.
   -- **** NOTE: THIS IS TARGET-DEPENDENT ****

   PROCEDURE_INFO_SIZE : constant := 32;


   --	    *************************
   --	    * 3. TYPE HISTORY_TABLE *
   --	    *************************

   -- HISTORY_TABLE is a private record containing the traceback table
   -- definition for each Ada code table. A pointer to the HISTORY_TABLE is
   -- held in the ADA_CODE_TABLE record. If the history is not held in the
   -- image, this pointer is either null or omitted.

   type HISTORY_TABLE is private;
   type REF_HISTORY_TABLE is access HISTORY_TABLE;

   -- Note: although this type is private, assembly language glue routines
   -- and the binder are allowed to assume knowledge of the details of the
   -- contents and layout of this record, as described in the private part. The
   -- The private declaration indicates and enforces the rule that the other
   -- parts of the Ada runtime, and in particular other parts of the exception
   -- manager, are NOT expected to mess with these structures (since they are
   -- target-dependent, and the rest of the exception manager is expected to
   -- be target-independent).


   --	    ******************
   --	    * 4. SUBPROGRAMS *
   --	    ******************

   -- **********************************************************************
   -- PROPAGATE_CONTEXT uses the supplied PROCEDURE_INFORMATION_RECORD to
   -- propagate the context to the calling procedure, modifying CONTEXT.
   -- In addition, NEW_MARKER is set to the appropriate value of the new
   -- clean up marker (i.e. to a value which will cause cleanup of all
   -- data in the called procedure whose frame is destroyed by the call).
   -- The cleanup marker may be left unchanged from its input value if
   -- there is no additional data to cleanup as a result of the propagation.

   procedure PROPAGATE_CONTEXT (INFO	   : PROCEDURE_INFORMATION_RECORD;
				CONTEXT    : REF_EXCEPTION_CONTEXT;
				NEW_MARKER : in out DATA_ADDRESS);

   -- This second version of PROPAGATE_CONTEXT is used to propagate the
   -- exception from a non-Ada frame. On entry, CONTEXT indicates that the
   -- PC is not in Ada code. Each target must consider whether it can
   -- propagate exceptions through non-Ada frames. Three options are
   -- supported:
   --	 (a) Propagation is not supported and the body aborts the program.
   --	     In this case, the call does not return.
   --	 (b) Propagation is not supported, but a stack traceback should be
   --	     generated first (if requested) followed by program abort.
   --	     In this case, the function should return FALSE.
   --	 (c) Propagation is supported. In this case, the function should
   --	     return TRUE and CONTEXT updated accordingly.

   function PROPAGATE_CONTEXT (CONTEXT : REF_EXCEPTION_CONTEXT)
						    return BOOLEAN;

   -- **********************************************************************
   -- GET_EXCEPTION_LOCATION obtains the current exception location from
   -- the exception context. It is typically just a field extraction.

   function GET_EXCEPTION_LOCATION (CONTEXT : REF_EXCEPTION_CONTEXT)
						    return CODE_ADDRESS;
   pragma INLINE (GET_EXCEPTION_LOCATION);


   -- **********************************************************************
   -- UPDATE_CONTEXT updates the exception context with the supplied location
   -- (which is the address of the handler which we are about to go to).
   -- It is typically just a field update.

   procedure UPDATE_CONTEXT (CONTEXT	     : REF_EXCEPTION_CONTEXT;
			     HANDLER_ADDRESS : CODE_ADDRESS);
   pragma INLINE (UPDATE_CONTEXT);


   -- **********************************************************************
   -- GET_EXCEPTION_FRAME_PTR obtains the current frame pointer value
   -- from the exception context. It is typically just a field extraction.
   -- This is used in processing an HTE_SET_MARKER entry in the exception
   -- tables (i.e. the offset in such an entry is relative to the value
   -- returned by this function).

   function GET_EXCEPTION_FRAME_PTR (INFO : PROCEDURE_INFORMATION_RECORD;
				     CONTEXT : REF_EXCEPTION_CONTEXT)
						    return DATA_ADDRESS;
   pragma INLINE (GET_EXCEPTION_FRAME_PTR);

   -- **********************************************************************
   -- EXITING_OUTER_FRAME is used to determine when propagation past the
   -- outermost frame of the Ada task or environment task has taken place,
   -- ie. it indicates task and program death. If this has occurred, a stack
   -- traceback is generated if the appropriate history option is set.

   function EXITING_OUTER_FRAME (CONTEXT : REF_EXCEPTION_CONTEXT)
							   return BOOLEAN;
   pragma INLINE (EXITING_OUTER_FRAME);
   
   -- **********************************************************************
   -- IS_ENVIRONMENT_TASK is used to determine whether the task for which
   -- EXITING_OUTER_FRAME is TRUE is an Ada task or the environment task.
   -- The Exception Manager invokes program shutdown actions if
   -- IS_ENVIRONMENT_TASK returns TRUE.

   function IS_ENVIRONMENT_TASK (CONTEXT : REF_EXCEPTION_CONTEXT)
							   return BOOLEAN;
   pragma INLINE (IS_ENVIRONMENT_TASK);
   
   -- **********************************************************************
   -- GET_ACTUAL_ECODE is used to covert what is potentially a pointer to
   -- an exception to an actual exception. ECODE can be a pointer to an
   -- exception if it names an imported exception.
   function GET_ACTUAL_ECODE (ECODE : in EXCEPTION_CODE) 
                                                     return EXCEPTION_CODE;
--   pragma INLINE (GET_ACTUAL_ECODE);
   
   -- **********************************************************************
   -- DUMP_PROCEDURE_INFO_RECORD is used in debug mode to dump the contents
   -- of a procedure information record. It should output a series of lines
   -- each preceded by 12 spaces, with one line feed at the end of the lines
   -- (i.e. NO extra blank line is generated before or after the output).

   procedure DUMP_PROCEDURE_INFO_RECORD (REC : PROCEDURE_INFORMATION_RECORD);


   -- **********************************************************************
   -- DUMP_EXCEPTION_CONTEXT is used in debug mode to dump the contents
   -- of the exception context record. It should output a series of lines
   -- each preceded by 12 spaces, with one line feed at the end of the lines
   -- (i.e. NO extra blank line is generated before or after the output).

   procedure DUMP_EXCEPTION_CONTEXT (CONTEXT : REF_EXCEPTION_CONTEXT);


   --	    ***********************************
   --	    * 4. PRIVATE DEFINITIONS OF TYPES *
   --	    ***********************************

private

   -- The following declaration defines the format of the context record which
   -- describes the current exception context. This is not the entire machine
   -- state, since it does not include registers which are not saved over call.

   subtype DISPLAY_LEVEL is INTEGER range 0..63;

   type DISPLAY_ARRAY is array (DISPLAY_LEVEL) of DATA_ADDRESS;
   type EXCEPTION_CONTEXT is record
      EIP         : CODE_ADDRESS;  
      EBP         : DATA_ADDRESS;
      ESP         : DATA_ADDRESS;
      TCB_POINTER : DATA_ADDRESS;
      DISPLAY     : DISPLAY_ARRAY;
   end record;

   for EXCEPTION_CONTEXT use record
      EIP         at  0 range 0 .. 31;
      EBP         at  4 range 0 .. 31;
      ESP         at  8 range 0 .. 31;
      TCB_POINTER at 12 range 0 .. 31;
      DISPLAY     at 16 range 0 .. 64*32-1;
   end record;
   

   -- The following declaration defines the format of the procedure information
   -- record which describes the static characteristics of a procedure for the
   -- purpose of restoring the calling context.

   type UINT_16 is mod 2**16;
   type PROCEDURE_INFORMATION_RECORD is record
      SAVED_DISPLAY_LEVEL : DISPLAY_LEVEL;
      EXPORT              : BOOLEAN;
      PARAMETER_SIZE      : UINT_16;
   end record;

   for PROCEDURE_INFORMATION_RECORD use record
      SAVED_DISPLAY_LEVEL at 0 range 0..7;
      EXPORT              at 1 range 0..7;
      PARAMETER_SIZE      at 2 range 0..15;
   end record;
   for PROCEDURE_INFORMATION_RECORD'SIZE use PROCEDURE_INFO_SIZE;

   -- The following declaration defines the format of the history table
   -- containing the stack traceback records. This need only be a dummy
   -- definition if the traceback records are not in the image, since in
   -- this case, the table cannot be referred to in each Ada code table,
   -- which is the only place in the common code where the type is used.

   type HISTORY_TABLE is record
      null;
   end record;

end System.RTS.TGT.Exceptions.TDEP;

--	    ***********************
--	    * 5. REVISION HISTORY *
--	    ***********************

--  6.01 20th December 1990 (BJD)
--	 Initial introduction to CMS v6 based on RS/6000 version.
--	 Remove target-dependency: context clauses, private type declarations,
--	 parameters to BUILD_PROCEDURE_INFO_RECORD, and revision history.

--  6.02 24th January 1991 (BJD)
--	 Add PROCEDURE_INFO_SIZE constant for use in rep clauses.
--	 Add EXITING_OUTER_FRAME and IS_ENVIRONMENT_TASK for controlling stack
--	 traceback generation and program termination.

-- Moved to TEMPLATE library 12th Mar 1991 (BJD)

--  6.02 13th March 1991 (BJD)
--	 Use "..." in target-dependent contexts according to rules for
--	 template files.

--  6.03 14th May 1991 (BJD)
--	 Add procedure information parameter to GET_EXCEPTION_FRAME_PTR as
--	 required by RS/6000.
--	 Add size clause for PROCEDURE_INFORMATION_RECORD as defensive
--	 programming.
--	 Move HISTORY_TABLE declaration here from EM_HIST to prevent EM_HIST
--	 body (or private part) from becoming target-dependent.

--  6.04  3rd November 1991 (BJD)
--	  Prefix all CRTS library unit names with ALSYS_

--  6.05 10th March 1993 (BJD)
--	 Add interface to allow propagation across non-Ada frames.

--  6.06 3rd June 1993 (BJD)
--	 Added procedure Update_Context for use by EM_Prop.Exception_Handle
--	 to update a context with the handler address prior to outputting the
--	 corresponding entry in the traceback, and prior to transferring
--	 control to that handler context.
