-- This is the Intel-specific version of the Generic_Elementary_Functions package
-- which goes straight to assembler and depends on the presence of a coprocessor.

package body Ada.Numerics.Generic_Elementary_Functions is

   PI    : constant := 3.14159_26535_89793_23846_26433_83279_50288_41972;
   TWO_PI          : constant :=   2.0 * PI;
   PI_ON_TWO       : constant :=   PI / 2.0;
   MINUS_PI_ON_TWO : constant := - PI / 2.0;
    
   DWORD : constant BOOLEAN := FLOAT_TYPE'BASE'SIZE=32;

   package FLOAT_MATH_LIB is
    
      function SQRT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the square root of X.
      -- The exception ARGUMENT_ERROR is raised if a negative argument is given.
    
      function EXP (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the exponential of X ( e to the X ).
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function "**" (X,Y : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes X to the power Y.
      -- The exception ARGUMENT_ERROR is raised if X < 0.0
      -- or ( X = 0.0 and Y < 0.0 ) .
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function LOG (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the napierian logarithm of X.
      -- The exception ARGUMENT_ERROR is raised if X is non-positive.
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function LOG10 (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the decimal logarithm of X.
      -- The exception ARGUMENT_ERROR is raised if X is non-positive.
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function SIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the sine of X expressed in radians.
    
      function COS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the cosine of X expressed in radians.
    
      function TAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the tangent of X expressed in radians.
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function COT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the cotangent of X expressed in radians.
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function ASIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the arc sine of X.
      -- The exception ARGUMENT_ERROR is raised if X not in -1.0 .. 1.0 .
    
      function ACOS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the arc cosine of X.
      -- The exception ARGUMENT_ERROR is raised if X not in -1.0 .. 1.0 .
    
      function ATAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the arc tangent of X.
    
      function SINH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the hyperbolic sine of X.
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function COSH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the hyperbolic cosine of X.
      -- The exception CONSTRAINT_ERROR is raised for too big result values.
    
      function TANH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
      -- Computes the hyperbolic tangent of X.
    
   private
    
      pragma INLINE(SQRT,EXP,"**",LOG,LOG10,SIN,COS,TAN,COT,
                    ASIN,ACOS,ATAN,SINH,COSH,TANH);
    
   end FLOAT_MATH_LIB;
 

   package body FLOAT_MATH_LIB is
   
      function SQRT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_SQRT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_SQRT, "_rts_math_DWORD_SQRT");
         function QWORD_SQRT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_SQRT, "_rts_math_QWORD_SQRT");
      begin    
         if DWORD then
            return DWORD_SQRT(X);
         else   
            return QWORD_SQRT(X);
         end if;
      end;
   
      function EXP (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_EXP (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_EXP, "_rts_math_DWORD_EXP");
         function QWORD_EXP (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_EXP, "_rts_math_QWORD_EXP");
      begin 
         if DWORD then
            return DWORD_EXP(X);
         else   
            return QWORD_EXP(X);
         end if;
      end;
   
      function "**" (X,Y : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_XY (X,Y : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_XY, "_rts_math_DWORD_XY");
         function QWORD_XY (X,Y : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_XY, "_rts_math_QWORD_XY");
      begin 
         if DWORD then
            return DWORD_XY(X,Y);
         else   
            return QWORD_XY(X,Y);
         end if;
      end;
   
      function LOG (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_LOG (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_LOG, "_rts_math_DWORD_LOG");
         function QWORD_LOG (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_LOG, "_rts_math_QWORD_LOG");
      begin 
         if DWORD then
            return DWORD_LOG(X);
         else   
            return QWORD_LOG(X);
         end if;
      end;
   
      function LOG10 (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_LOG10 (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_LOG10, "_rts_math_DWORD_LOG10");
         function QWORD_LOG10 (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_LOG10, "_rts_math_QWORD_LOG10");
      begin 
         if DWORD then
            return DWORD_LOG10(X);
         else   
            return QWORD_LOG10(X);
         end if;
      end;
   
      function SIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_SIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_SIN, "_rts_math_DWORD_SIN");
         function QWORD_SIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_SIN, "_rts_math_QWORD_SIN");
      begin 
         if DWORD then
            return DWORD_SIN(X);
         else   
            return QWORD_SIN(X);
         end if;
      end;
   
      function COS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_COS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_COS, "_rts_math_DWORD_COS");
         function QWORD_COS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_COS, "_rts_math_QWORD_COS");
      begin 
         if DWORD then
            return DWORD_COS(X);
         else   
            return QWORD_COS(X);
         end if;
      end;
   
      function TAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_TAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_TAN, "_rts_math_DWORD_TAN");
         function QWORD_TAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_TAN, "_rts_math_QWORD_TAN");
      begin 
         if DWORD then
            return DWORD_TAN(X);
         else   
            return QWORD_TAN(X);
         end if;
      end;
   
      function COT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_COT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_COT, "_rts_math_DWORD_COT");
         function QWORD_COT (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_COT, "_rts_math_QWORD_COT");
      begin 
         if DWORD then
            return DWORD_COT(X);
         else   
            return QWORD_COT(X);
         end if;
      end;
   
      function ASIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_ASIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_ASIN, "_rts_math_DWORD_ASIN");
         function QWORD_ASIN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_ASIN, "_rts_math_QWORD_ASIN");
      begin 
         if DWORD then
            return DWORD_ASIN(X);
         else   
            return QWORD_ASIN(X);
         end if;
      end;
   
      function ACOS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_ACOS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_ACOS, "_rts_math_DWORD_ACOS");
         function QWORD_ACOS (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_ACOS, "_rts_math_QWORD_ACOS");
      begin 
         if DWORD then
            return DWORD_ACOS(X);
         else   
            return QWORD_ACOS(X);
         end if;
      end;
   
      function ATAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_ATAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_ATAN, "_rts_math_DWORD_ATAN");
         function QWORD_ATAN (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_ATAN, "_rts_math_QWORD_ATAN");
      begin 
         if DWORD then
            return DWORD_ATAN(X);
         else   
            return QWORD_ATAN(X);
         end if;
      end;
   
      function SINH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_SINH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_SINH, "_rts_math_DWORD_SINH");
         function QWORD_SINH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_SINH, "_rts_math_QWORD_SINH");
      begin 
         if DWORD then
            return DWORD_SINH(X);
         else   
            return QWORD_SINH(X);
         end if;
      end;
   
      function COSH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_COSH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_COSH, "_rts_math_DWORD_COSH");
         function QWORD_COSH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_COSH, "_rts_math_QWORD_COSH");
      begin 
         if DWORD then
            return DWORD_COSH(X);
         else   
            return QWORD_COSH(X);
         end if;
      end;
   
      function TANH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE is
         function DWORD_TANH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, DWORD_TANH, "_rts_math_DWORD_TANH");
         function QWORD_TANH (X : FLOAT_TYPE'BASE) return FLOAT_TYPE'BASE;
         pragma IMPORT (ASSEMBLER, QWORD_TANH, "_rts_math_QWORD_TANH");
      begin 
         if DWORD then
            return DWORD_TANH(X);
         else   
            return QWORD_TANH(X);
         end if;
      exception
         when others =>
            return 0.0;
      end;
   
   end FLOAT_MATH_LIB;
   
   function SQRT (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
       FLOAT_MATH_LIB.SQRT;
   
   function LOG (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if X < 0.0 then
         raise ARGUMENT_ERROR; -- ARM A.5.1(22)
      elsif X=0.0 then
         raise CONSTRAINT_ERROR; -- ARM A.5.1(29)
      else
         return FLOAT_MATH_LIB.LOG (X);
      end if;
   end LOG;
   
   function LOG (X, BASE   : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if BASE > 0.0 then
         if BASE /= 1.0 then
            return FLOAT_MATH_LIB.LOG (X) / FLOAT_MATH_LIB.LOG (BASE);
	 else
	    raise CONSTRAINT_ERROR; -- ARM A.5.1(29)
	 end if;
      else
         raise ARGUMENT_ERROR; -- ARM A.5.1(22)
      end if;
   end LOG;
   
   function EXP (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.EXP;
   
   function "**" (Left, Right : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB."**";
   
   function SIN (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.SIN;
   
   type EINT is range -2..7;
   LARGEARG : constant EINT := -2;
   NOTEXACT : constant EINT := -1;
   function EIGHTH (X, CYCLE  : FLOAT_TYPE'BASE ) return EINT is
      N:INTEGER;
      Q:FLOAT_TYPE'BASE:=CYCLE/8.0;
   begin
      N:=INTEGER(X/Q);
      if(FLOAT_TYPE'BASE(N)*Q=X) then
         return EINT(N mod 8);
      else 
         return NOTEXACT;
      end if;
   exception
      -- V. large X could cause this
      when CONSTRAINT_ERROR =>
         return LARGEARG;
   end EIGHTH;      

   function SIN (X, CYCLE  : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         case EIGHTH (X,CYCLE) is
            when 0 =>
               return 0.0;
            when 2 =>
               return 1.0;
            when 4 => 
               return 0.0;
            when 6 =>
               return -1.0;
            when LARGEARG =>
               return 0.0;
            when others =>
               return FLOAT_MATH_LIB.SIN (X / CYCLE * TWO_PI);
         end case;
      else
         raise ARGUMENT_ERROR;
      end if;
   end SIN;
       
   function COS (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.COS;
   
   function COS (X, CYCLE  : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         case EIGHTH (X,CYCLE) is
            when 0 =>
               return 1.0;
            when 2 =>
               return 0.0;
            when 4 => 
               return -1.0;
            when 6 =>
               return 0.0;
            when LARGEARG =>
               return 0.0;
            when others =>
               return FLOAT_MATH_LIB.COS (X / CYCLE * TWO_PI);
         end case;
      else
         raise ARGUMENT_ERROR;
      end if;
   end COS;
   
   function TAN (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.TAN;
   
   function TAN (X, CYCLE  : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         case EIGHTH (X,CYCLE) is
            when 0|4 =>
               return 0.0;
            when 1|5 =>
               return 1.0;
            when 2|6 =>
               raise Constraint_Error;
            when 3|7 =>
               return -1.0;
            when LARGEARG =>
               return 0.0;
            when NOTEXACT =>
               return FLOAT_MATH_LIB.TAN (X / CYCLE * TWO_PI);
         end case;
      else
         raise ARGUMENT_ERROR;
      end if;
   end TAN;
   
   function COT (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.COT;
   
   function COT (X, CYCLE  : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         case EIGHTH (X,CYCLE) is
            when 0|4 =>
               raise Constraint_Error;
            when 1|5 =>
               return 1.0;
            when 2|6 =>
               return 0.0;
            when 3|7 =>
               return -1.0;
            when LARGEARG =>
               return 0.0;
            when NOTEXACT =>
               return FLOAT_MATH_LIB.COT (X / CYCLE * TWO_PI);
         end case;
      else
         raise ARGUMENT_ERROR;
      end if;
   end COT;
   
   
   function ARCSIN (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.ASIN;
   
   function ARCSIN (X, CYCLE  : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         return FLOAT_MATH_LIB.ASIN (X) * CYCLE / TWO_PI;
      else
         raise ARGUMENT_ERROR;
      end if;
   end ARCSIN;
   
   function ARCCOS (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.ACOS;
   
   function ARCCOS (X, CYCLE  : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         return FLOAT_MATH_LIB.ACOS (X) * CYCLE / TWO_PI;
      else
         raise ARGUMENT_ERROR;
      end if;
   end ARCCOS;
   
   function ARCTAN (Y         : FLOAT_TYPE'BASE;
                         X : FLOAT_TYPE'BASE := 1.0) return FLOAT_TYPE'BASE is
   begin
      if X > 0.0 then
         return FLOAT_MATH_LIB.ATAN (Y / X);
      elsif X < 0.0 then
         if Y >= 0.0 then
            return (FLOAT_MATH_LIB.ATAN (Y / X) + PI);
         elsif Y < 0.0 then
            return (FLOAT_MATH_LIB.ATAN (Y / X) - PI);
         end if;
      else
         if Y > 0.0 then
            return PI_ON_TWO;
         elsif Y < 0.0 then
            return - PI_ON_TWO;
         else
            raise ARGUMENT_ERROR;
         end if;
      end if;
   end ARCTAN;
   
   function ARCTAN (Y         : FLOAT_TYPE'BASE;
                         X : FLOAT_TYPE'BASE := 1.0;
                         CYCLE     : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if CYCLE > 0.0 then
         return ARCTAN (Y, X) * CYCLE / TWO_PI;
      else
         raise ARGUMENT_ERROR;
      end if;
   end ARCTAN;
   
   function ARCCOT (X : FLOAT_TYPE'BASE;
                         Y         : FLOAT_TYPE'BASE := 1.0) return FLOAT_TYPE'BASE is
   begin
      return ARCTAN (Y,X);
      if X /= 0.0 then
         return FLOAT_MATH_LIB.ATAN (Y / X);
      else
         if Y > 0.0 then
            return PI_ON_TWO;
         elsif Y < 0.0 then
            return MINUS_PI_ON_TWO;
         else
            raise ARGUMENT_ERROR;
         end if;
      end if;
   end ARCCOT;
   
   function ARCCOT (X : FLOAT_TYPE'BASE;
                         Y         : FLOAT_TYPE'BASE := 1.0;
                         CYCLE     : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      return ARCTAN (Y,X,CYCLE);
      if CYCLE > 0.0 then
         if X /= 0.0 then
            return FLOAT_MATH_LIB.ATAN (Y / X) * CYCLE / TWO_PI;
         else
            if Y > 0.0 then
               return PI_ON_TWO;
            elsif Y < 0.0 then
               return MINUS_PI_ON_TWO;
            else
               raise ARGUMENT_ERROR;
            end if;
         end if;
      else
         raise ARGUMENT_ERROR;
      end if;
   end ARCCOT;
   
   function SINH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.SINH;
   
   function COSH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE renames
     FLOAT_MATH_LIB.COSH;
   
   function TANH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      return FLOAT_MATH_LIB.TANH(X);
   end TANH;
      
   
   function COTH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if X /= 0.0 then
         return 1.0 / FLOAT_MATH_LIB.TANH (X);
      else
         raise CONSTRAINT_ERROR;
      end if;
   exception 
      when CONSTRAINT_ERROR =>
         return 0.0;
   end COTH;
   
   function ARCSINH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      return FLOAT_MATH_LIB.LOG (X + SQRT (X*X + 1.0));
   end ARCSINH;
   
   function ARCCOSH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
   begin
      if X >= 1.0 then
         return FLOAT_MATH_LIB.LOG (X + SQRT (X*X - 1.0));
      else
         raise ARGUMENT_ERROR; -- ARM A.5.1 (26)
      end if;
   end ARCCOSH;
   
   function ARCTANH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
      ABSX : constant FLOAT_TYPE'BASE := abs X;
   begin
      if ABSX < 1.0 then 
         return FLOAT_MATH_LIB.LOG ((1.0 + X) / (1.0 - X)) / 2.0;
      elsif ABSX=1.0 then
         raise CONSTRAINT_ERROR; -- ARM A.5.1(33)
      else 
         raise ARGUMENT_ERROR; -- ARM A.5.1(24)
      end if;
   end ARCTANH;
   
   function ARCCOTH (X : FLOAT_TYPE'BASE ) return FLOAT_TYPE'BASE is
      ABSX : constant FLOAT_TYPE'BASE := abs X;
   begin
      if ABSX > 1.0 then 
         return FLOAT_MATH_LIB.LOG ((X + 1.0) / (X - 1.0)) / 2.0;
      elsif ABSX=1.0 then
         raise CONSTRAINT_ERROR; -- ARM A.5.1(33)
      else
         raise ARGUMENT_ERROR; -- ARM A.5.1(24)
      end if;
   end ARCCOTH;
   
   
end Ada.Numerics.Generic_Elementary_Functions;
