-- $Source: /home/dsd/d/aa/monoBANK/AA/system.rts.tgt.basic_io.spc,v $
-- $Revision: 1.3 $ -- $Date: 95/09/19 15:18:29 $ -- $Author: akira $

with Ada.IO_Exceptions;
with System;
package System.RTS.Low_Level_IO is  --| IO operations underlying sequential
                                       --| direct, and stream IO's
    pragma Suppress( Elaboration_Check );

    --| OVERVIEW
    --|
    --| Basic_IO is the low level interface for Sequential, Direct and
    --| Stream I/O.  It contains the basic operations for creating, opening,
    --| reading, writing, and the various queries (e.g. end-of-file) that
    --| are needed for sequential, direct and stream I/O.  The amount of
    --| data to read or write at any one time is expressed in Number of Bits.
    --| The Number of Bits depend upon the type of element to be operated
    --| upon. The operations distinguish among different I/O via an
    --| enumeration type.
    --| Basic_IO also defines a File Type that describes an external file.

    type IO_Kind is (Text_IO, Direct_IO, Sequential_IO, Stream_IO);
         --| Indicates the kind of IO being used.

    type File_Mode is (In_File, Out_File, Append_File, InOut_File);
    subtype Char_File_Mode is File_Mode range In_file..Append_File;
         --| IO files can be of mode In_File (read), Out_File (write),
	 --| or InOut_File (read/write).

    subtype Address_Type is System.Address;
         --| System address, used to designate location in which to read
         --| or write elements.

    type File_Type is private;

----------------------------------------------------------------------

    procedure Create (
        File      : in out File_Type;   --| File to be associated with new
                                        --|   external file.
        Mode      : in File_Mode;       --| Mode of new external file.
        Name      : in String;          --| Name of new external file.
        Form      : in String;          --| Form of new external file.
        Kind      : in IO_Kind;         --| Kind of IO used.
        NumUnit   : in Integer          --| Number of bits in element type.
        );

    --| OVERVIEW
    --| Establishes new external file with given name and form, and
    --| associates the external file with the given file.  Leaves the
    --| external file open.  If no name is given, a temporary file is
    --| created.  If no mode is given, the default mode for sequential
    --| and stream IO is Out_File, and the default mode for direct IO
    --| is InOut_File.
    --| This routine requires that the given file is not already open.
    --| If it is already open, Status_Error is raised.  If the string
    --| given as name does not allow the identification of an external
    --| file, Name_Error is raised.  If, for the specified mode, the
    --| environment does not support creation of an external file with
    --| the given name (in the absence of Name_Error) and form, Use_Error
    --| is raised.


----------------------------------------------------------------------

    procedure Open (
        File      : in out File_Type;   --| File to be associated with
                                        --|   external file.
        Mode      : in File_Mode;       --| Mode of external file.
        Name      : in String;          --| Form of external file.
        Form      : in String;          --| Form of external file.
        Kind      : in IO_Kind;         --| Kind of IO.
        NumUnit   : in Integer          --| Number of bits in element type.
        );

    --| OVERVIEW
    --| Associates the given file with an external file having the given
    --| name and form, and sets the current mode of the given file to
    --| the given mode. The given file is left open.
    --| This routine requires that the given file is not already open.  If
    --| the given file is already open, Status_Error is raised.  If the
    --| string given as name does not allow the identification of an
    --| external file, Name_Error is raised.  If the environment does
    --| not support opening of an external file with the given name and
    --| mode, Use_Error is raised.


----------------------------------------------------------------------

    procedure Close (
        File      : in out File_Type    --| File associated with external file
        );

    --| OVERVIEW
    --| Severs the association between the given file and its associated
    --| external file.  The given file is left closed.  For temporary
    --| files, the file is deleted upon close.
    --| If the given file is not open, Status_Error is raised.

 ----------------------------------------------------------------------

    procedure Delete (
        File      : in out File_Type    --| File whose external file is to be
                                        --|   deleted.
        );

    --| OVERVIEW
    --| Deletes the external file associated with the given file. The
    --| given file is closed, and the external file ceases to exist.
    --| If the given file is not open, Status_Error is raised.  If the
    --| environment does not support deletion of the external file
    --| associated with the given file, Use_Error is raised.

------------------------------------------------------------------

    procedure Switch_Mode (
        File      : in out File_Type;   --| File associated with external file
                                        --|   whose mode is to be switched
        Mode  : in File_Mode        --| New mode of external file.
        );

    --| OVERVIEW
    --| The mode of the external file associated with the given file is
    --| switched to a new mode.  If the mode cannot be switched,
    --| Internal_Error is raised.

------------------------------------------------------------------

    procedure Reset (
        File      : in out File_Type    --| File associated with external file
                                        --|   which is to be reset.
        );

    --| OVERVIEW
    --| Resets the given file so that reading from or writing to it
    --| elements can be restarted from the beginning of the file; in
    --| particular, for direct access this means that the current index is
    --| set to one.  Raises Internal_Error if this cannot be done.

------------------------------------------------------------------

    function Name (
      File      : in File_Type         --| File associated with external
                                       --|   file whose name is to be returned.
        ) return String;               --| Name of external file.

    --| OVERVIEW
    --| Returns the name of the external file associated with the given
    --| file.  If the given file is not open, Status_Error is raised.

--------------------------------------------------------------------------

    function Form (
        File      : in File_Type        --| File associated with external file
                                        --|   whose form is to be returned.
        ) return String;                --| Form of external file.

    --| OVERVIEW
    --| Returns the form string of the external file associated with the
    --| given file.  If the given file is not open, Status_Error is raised.

--------------------------------------------------------------------------

    function Read (
        File      : in File_Type;       --| File whose associated external
                                        --|   file bits are to be read from.
        Where     : in Address_Type;    --| Where to put the bits read.
        Amount    : in Natural          --| How many bits to read.
        ) return Natural;

    --| OVERVIEW
    --| The specified number of bits are read from the external file
    --| associated with the given file.  The bits are read from the
    --| current position in the external file.  The current index
    --| into the file is updated to reflect the read if the file is
    --| a direct IO file.  The number of bits read is returned.
    --| This procedure requires that the correct number of bits can be read
    --| from the external file.  If the requested number of bits cannot
    --| be read, Data_Error is raised.  If some other error occurs,
    --| Internal_Error is raised.

    function Input_Available (
        File      : in File_Type) return Boolean;
    -- Perform a non-blocking test on the file

----------------------------------------------------------------------

    function Write (
        File        : in File_Type;     --| File associated with external file
                                        --|   where bits are going to be written
        Where       : in Address_Type;  --| Location of bits to be written
        Amount      : in Natural        --| Number of bits to be written.
        ) return Natural;

    --| OVERVIEW
    --| The specified number of bits are written to the external file
    --| associated with the given file.  The bits are written to the
    --| current position in the external file.  The current index
    --| into the file is updated to reflect the writing if the file is
    --| a direct IO file.  The number of bits written is returned.
    --| This procedure requires that the correct number of bits can be
    --| written to the external file.  If this is not possible,
    --| Internal_Error is raised.

----------------------------------------------------------------------

    procedure Set_Index (
        File      : in File_Type;       --| File associated with external file
                                        --|   whose index is to be set.
        To        : in integer          --| Index number that current index
                                        --|   is to be set to.
        );

    --| OVERVIEW
    --| Sets the index of the specified file to the given index value
    --| (which may exceed the size of the file).  This sets the current
    --| position in the file to index number of elements.
    --| This procedure requires that the current position in the file
    --| can be changed to the desired index.  If this cannot happen,
    --| Internal_Error is raised.

------------------------------------------------------------------

    function Index (
        File      : in File_Type        --| File associated with external file
                                        --|   whose index is to be returned.
        ) return integer;               --| Returns current index into file
                                        --| (current element)

    --| OVERVIEW
    --| Returns the current index of the given file.
    --| If the given file is not open, Status_Error is raised.

----------------------------------------------------------------------

    function Size (
        File      : in File_Type        --| File associated with external file
                                        --|   whose size is to be returned.
        ) return integer;               --| Size of external file in bits.

    --| OVERVIEW
    --| The size of the external file associated with the given file is
    --| returned (in bits).

----------------------------------------------------------------------

    function End_Of_File (
        File      : in File_Type        --| File associated with external file
                                        --|   whose end status to be returned.
        ) return Boolean;               --| True if at end of file.

    --| OVERVIEW
    --| Returns TRUE if no more bytes of data can be read from the given file;
    --| otherwise returns false.

----------------------------------------------------------------------

    procedure OPEN_STANDARD_INPUT (
        FILE     : in out FILE_TYPE     --| File which is to be associated
                                        --|   with external file.
        );
    procedure OPEN_STANDARD_OUTPUT (
        FILE     : in out FILE_TYPE     --| File which is to be associated
                                        --|   with external file.
        );

    procedure OPEN_STANDARD_ERROR (
        FILE     : in out FILE_TYPE     --| File which is to be associated
                                        --|   with external file.
        );

    procedure Flush (FILE : in FILE_TYPE);

    procedure Set_Console (FILE : in FILE_TYPE;Char_Mode : in Boolean);

--------------------------------------------------------------------------

   --| Exceptions

   Status_Error : exception renames Ada.IO_Exceptions.Status_Error;
   Mode_Error   : exception renames Ada.IO_Exceptions.Mode_Error;
   Name_Error   : exception renames Ada.IO_Exceptions.Name_Error;
   Use_Error    : exception renames Ada.IO_Exceptions.Use_Error;
   Device_Error : exception renames Ada.IO_Exceptions.Device_Error;
   End_Error    : exception renames Ada.IO_Exceptions.End_Error;
   Data_Error   : exception renames Ada.IO_Exceptions.Data_Error;

private

    type File_Record;
    type File_Type is access File_Record;

end System.RTS.Low_Level_IO;
