-- == %filter "-- "
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-- This unpublished work is protected both as a proprietary work and under
-- the Universal Copyright Convention and the US Copyright Act of 1976.  Its
-- distribution and access are limited only to authorized persons.  Copyright
-- (C) Alsys.  Created 1990, initially licenced 1990.  All rights reserved.
-- Unauthorized use (including use to prepare other works), disclosure,
-- reproduction or distribution may violate national criminal law.
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

------------------------------------------------------------------------------
-- PACKAGE SPECIFICATION : System.Rts.Types
--
-- This module defines the representation-sensitive types used throughout the
-- common runtime. These types are defined as derivatives of target-dependent
-- types defined elsewhere, primarily SYSTEM (for address-related types) and
-- TARGET_INTEGERS (for integer-related types). Thus the parent type in each
-- derived type definition is target-dependent. Also included are special
-- constants of these types.
------------------------------------------------------------------------------

with SYSTEM;
with UNCHECKED_CONVERSION;
with System.RTS.Tgt_INTEGERS; use System.RTS.Tgt_INTEGERS;

pragma ELABORATE (SYSTEM);
package System.Rts.Types is

   pragma SUPPRESS (ELABORATION_CHECK);

   --	    *********************
   --	    * TABLE OF CONTENTS *
   --	    *********************

   --	      1. Global Constants
   --	      2. Common Integer Types
   --		 2.1 Enumeration Support Types
   --		 2.2 Integer Support Types
   --		 2.3 Heap Support Types
   --		 2.4 Exception Support Types
   --		 2.5 Tasking Support Types
   --	      3. Common Code Address Type
   --	      4. Common Data Address Type

   -- ------------------------------------------
   -- Package renamings to remove ALSYS_ prefix:
   -- ------------------------------------------
   package TARGET_INTEGERS renames System.RTS.Tgt_INTEGERS;
   -- ------------------------------------------

   --	    ***********************
   --	    * 1. Global Constants *
   --	    ***********************

   --	None at present.

   --	    ***************************
   --	    * 2. Common Integer Types *
   --	    ***************************

   -- Each target must complete the following declarations using the
   -- appropriate target-dependent type. These types will be used by
   -- the common code in representation-sensitive contexts such as the
   -- Root/RTS subprogram parameter types.

   -- -----------------------------
   -- 2.1 Enumeration Support Types
   -- -----------------------------

   -- Type for number of elements in an enumeration type.
   -- This is used in the Root/RTS interface for passing the number of
   -- elements in enum tables, and for indicating position numbers.

   type ENUM_COUNT is mod 2**16;

   -- Types to describe the representation of enum types with user-defined
   -- enumeration rep clause values, including size and signed/unsigned.
   -- Each target must complete all the type declarations even if some of
   -- the representations are not supported. In this case, "missing" types
   -- should be completed using any of the supported sizes.
   --
   -- Note that the Root chooses the RTS routine to call
   -- on the basis of the number of Storage Units in the enum object size
   -- and whether the representation is signed or unsigned. For combinations
   -- which are unsupported, the corresponding RTS routine will not be
   -- called by the Root and so the code will be eliminated at Bind time.
   -- There is no longer any need for filtering in the Root or in the RTS.
   --
   -- Examples:
   -- (a) No 32-bit type:
   --	  type ENUM_1SU is new INTEGER_8;
   --	  type ENUM_2SU is new INTEGER_16;
   --	  type ENUM_4SU is new INTEGER_16;
   -- (b) No 16-bit type:
   --	  type ENUM_1SU is new INTEGER_8;
   --	  type ENUM_2SU is new INTEGER_32;
   --	  type ENUM_4SU is new INTEGER_32;
   -- (c) No unsigned type:
   --	  type ENUM_1SU_UNSIGNED is new INTEGER_8;
   --	  type ENUM_2SU_UNSIGNED is new INTEGER_16;

   type ENUM_1SU is mod 2**8;
   type ENUM_2SU is mod 2**16;
   type ENUM_4SU is mod 2**32;
   type ENUM_1SU_UNSIGNED is mod 2**8;
   type ENUM_2SU_UNSIGNED is mod 2**16;

   -- Type of string length field in the enum image table. (Note that the
   -- characters themselves are assumed to be represented as STRING).
   type ENUM_IMAGE_LENGTH is new UNSIGNED_8;

   -- Type of elements in the enum index table
   type ENUM_INDEX_ELEM is mod 2**16;

   -- -------------------------
   -- 2.2 Integer Support Types
   -- -------------------------

   -- This type should have at least 16 bits of range, and should be the
   -- most efficient integer type. It is used for all general counters,
   -- such as the number of tasks on a queue.
   type COUNTER is new INTEGER_32;

   -- The predefined integer type with the largest range on the target.
   type LONGEST_INTEGER is new INTEGER_32;

   -- ----------------------
   -- 2.3 Heap Support Types
   -- ----------------------

   -- In the Small_Objects heap, this type is used to hold the offset to
   -- the next free list item. For storage economy, this should be the
   -- shortest unsigned integer type.
   type FREE_LIST_ELEM is new UNSIGNED_8;

   -- ---------------------------
   -- 2.4 Exception Support Types
   -- ---------------------------

   -- This type is used to hold an exception code. It is used by the Root
   -- as the container size for the anonymous raise implicit object.
   -- Exception code value assigned by binder. 32 bits is the appropriate
   -- choice on most targets, but on some targets it may be possible to
   -- compress these codes to 16 bits.
--    == % if TARGET_I80386, TARGET_RIOS
   type EXCEPTION_CODE is new INTEGER_32;
--    == % end if

   type CODE_OFFSET is new INTEGER_32;
   -- Code offset in bytes from the start of a given block of Ada
   -- code referenced in the Ada Code Table.

   type PROC_OFFSET is new INTEGER_32;
   -- Code offset in bytes from the start of the relevant procedure.

   type FRAME_OFFSET is new INTEGER_32;
   -- Frame offset in the handler table, in bytes

   type SCOPE_TABLE_SUBSCRIPT is new INTEGER_32;
   -- Subscript for referencing SCOPE_TABLE (See EM_TYPES)

   type HANDLER_TABLE_SUBSCRIPT is new INTEGER_32;
   -- Subscript for referencing HANDLER_TABLE (See EM_TYPES)

   type ENTRY_TABLE_SUBSCRIPT is new INTEGER_32;
   -- Subscript for referencing ENTRY_TABLE or PROCEDURE_TABLE (See EM_TYPES)


   -- -------------------------
   -- 2.5 Tasking Support Types
   -- -------------------------

   -- These types define the types of the fields Preemptable and Abortable
   -- in the common TCB. In view of the special requirements attached to these
   -- fields, eg. indivisible increment/decrement, their representation is
   -- target-dependent.
   type PREEMPTABLE_TYPE is new INTEGER_8;

   type ABORTABLE_TYPE is new INTEGER_32;

   -- -----------------------------------
   -- 2.6 Interrupt Entries Support Types
   -- -----------------------------------

   -- Some of these definitions need to be made visible to the user via
   -- the Interrupt_manager package. It is expected that this would be done
   -- by renames in the target's interrupt manager package.

   -- ID of a handler. This is used internally to index tables and identify
   -- a user installed handler. Here, we choose to map onto COUNTER.
   subtype HANDLER_ID is COUNTER;

   -- Maximum number of handlers which can be simultaneously attached. Note
   -- that this constant must fit in the range of HANDLER_ID.
--    == % if TARGET_BARE_I80386, TARGET_MS_DOS, TARGET_OS2
--    MAX_HANDLERS : constant := 16;
--    == % elsif TARGET_LYNX
--    MAX_HANDLERS : constant := 64;
--    == % elsif TARGET_UNIX, TARGET_NT
   MAX_HANDLERS : constant := 32;
--    == % end if

   -- Again, a simple identifier for buffers. This is used internally to
   -- identify buffers. The range must be big enough to accommodate the
   -- largest number of buffers that the user can request at Initialization.
   subtype BUFFER_NUMBER is COUNTER;

   -- This is the type that the atomic instructions will use in the
   -- implementation of interrupt entries. This package requires concurrent
   -- updates (increment, decrement and swap) of counters and status flags
   -- and this defines the most convenient type for these operations.
   type ATOM_TYPE is new TARGET_INTEGERS.INTEGER_32;

   -- Type used to allocate interrupt stack and for the param area in the
   -- interrupt rendezvous.
   type STACK_ELEM_TYPE is new TARGET_INTEGERS.INTEGER_8;

   --	    *******************************
   --	    * 3. Common Code Address Type *
   --	    *******************************

   -- Code address type. The operations on code addresses are assignments,
   -- comparisons of addresses (used in the exception manager for the
   -- binary search used to locate the proper procedure in a code block,
   -- and for the range tests use to locate the proper code block and to
   -- locate the correct handler in a block). Finally we can add and subtract
   -- integer values from a code address or subtract one code address from
   -- another to get an integer difference in System.Offset units.

   -- Note that we never read the code in target-independent modules since
   -- in general, the code may not be readable.

   -- If Code_Address is derived from System.Address, then the relational and
   -- arithmetic operations on code addresses are those inherited from
   -- package System. Otherwise, specific operators must be defined here.

   type CODE_ADDRESS is new SYSTEM.ADDRESS;
   NULL_CODE_ADDRESS : constant CODE_ADDRESS :=
					    CODE_ADDRESS(SYSTEM.NULL_ADDRESS);

   --	    *******************************
   --	    * 4. Common Data Address Type *
   --	    *******************************

   -- Data address type. The operations on data addresses are assignments,
   -- addition of a constant to an address to get another address, and the
   -- extraction of values at a given address. If Data_Address is defined
   -- as System.Address then the arithmetic operations on data addresses are
   -- those defined in package System and relational operations on data
   -- addresses are those defined in package Stack_Extension. Otherwise,
   -- specific operators must be defined here.

   type DATA_ADDRESS is new SYSTEM.ADDRESS;
   NULL_DATA_ADDRESS : constant DATA_ADDRESS :=
					    DATA_ADDRESS(SYSTEM.NULL_ADDRESS);

--II   DATA_ADDRESS_SIZE : constant := SYSTEM.ADDRESS_SIZE;
   DATA_ADDRESS_SIZE : constant := 32;
                   
   -- ----------------------------------------------------------------------
   -- Special values of DATA_ADDRESS used by the tasking kernel and heap mgr
   -- ----------------------------------------------------------------------

   -- (a) IMPOSSIBLE_TCB_ADDRESS:
   --	 This field should be set to a DATA_ADDRESS value which cannot be a
   -- valid TCB address. NOTE that the equivalent value of the null access
   -- value is NOT permitted. This special non-null, non-valid-TCB address is
   -- used to disable the tasking scheduler cache when time-slicing is
   -- enabled.

   function INTEGER_32_TO_ADDRESS is new UNCHECKED_CONVERSION
      (INTEGER_32, DATA_ADDRESS );

   -- On the 386, stacks grow from high addresses (0x7FFFFFFF) down
   -- towards zero. 

   IMPOSSIBLE_TCB_ADDRESS : constant DATA_ADDRESS := INTEGER_32_TO_ADDRESS (1);

--    == % if TARGET_NT
   -- These are not used
--    == % else
--    -- Special frame pointer values
--    ENV_TASK_FP  : constant DATA_ADDRESS := INTEGER_32_TO_ADDRESS(-2);
--    TASK_BODY_FP : constant DATA_ADDRESS := INTEGER_32_TO_ADDRESS (1);
--    == % end if

end System.Rts.Types;

-- Modification History:

--  1.01  15 May 1990, 08:01 (RBKD)
--	     Initial version

--  6.01  26th September 1990, (BJD)
--	     Initial introduction to Shared RTS libraries

--  6.02  12th October 1990 (BJD)
--	     Byte->Unsigned_8, Word->Unsigned_16, added Extended_string,
--	  Longest_Integer, Null Data and Code address, Older on data
--	  addresses, and DEBUG constant. Used System.Offset instead of
--	  Integer_32 for offsets.

--  6.03  9th November 1990 (BJD)
--	      Re-engineered Rt_Types to define target-dependent types and
--	  target-independent subtypes.
--
--	      Added Unsigned_32, removed Refs to Float (not used), removed
--	  address arithmetic and relational operators (use the ones in
--	  System and Stack_Extension), used System.Fetch_From_Address and
--	  Assign_To_Address for Get and Put operations taking Data_Addresses,
--	  assumed Unchecked_Conversion for converting Data_Addresses from/to
--	  access types.

--  6.04  13th November 1990 (BJD)
--	      Changed target-independent subtypes to derived types to ensure
--	  compile-time detection of type mixing. Introduced package TARGET_
--	  INTEGERS to get round RM [3.4:15]. Added zero collection sizes
--	  for access types to suppress heap initialisation calls. Removed
--	  'use' of System as recommended by coding standards.

--  6.05  19th November 1990 (BJD)
--	  Add initial filter directive

--  6.06  27th November 1990 (APC)
--	      Added type REGISTER. Removed type EXTENDED_STRING which was
--	  temporary.

--  6.07  7th  December 1990 (SDF)
--	      Commented out 'STORAGE_SIZE clauses until compiler supports them
--	  (currently such clauses cause an expander error or crash)

--  6.08  9th January 1991 (BJD)
--	  Removed type REGISTER (no longer required by ERROR_IO). Extracted
--	  Target_Integers into a separate package to improve encapsulation.
--	  Added special constant Data_Addresses used by Tasking Kernel.
--	  Removed REF_ types (not used).

--  6.09  23rd January 1991 (BJD)
--	  Correct filter tag name.
--	  Remove DEBUG Boolean (local Booleans or filtering preferred).
--	  Add type COUNTER for all counting variables.
--	  Move rep-sensitive integer types from EM_TYPES to here.

--  6.10  29th January 1991 (CEB)
--	  Added an explanation for the task hierarchy termination values
--	  used on 386 as an example of their selection.

--  6.11  31st January 1991 (APC)
--	  Added HEAP_LIST_TERMINATOR which is used by FREE_LIST.
--	  Removed GET and PUT, which aren't used, and can't be inlined anyway
--	  because of the dependency order.
--	  Added null body to bring in line with coding standards.

--  6.12  11th February 1991 (DMW)
--	  Added special values of DATA_ADDRESS used by tasking kernel for 370.

--  6.13  18th February 1991 (CEB)
--	  Made exception codes 16 bits for 386 and preempt count 8 bits
--	  used filtering to protect 370 from exception code change.

--  6.14  7th March 1991 (APC)
--	  Changed the value of HEAP_LIST_TERMINATOR since the heap assumes
--	  this is even.

--  6.15  12th March 1991 (BJD)
--	  Add IMPOSSIBLE_TCB_ADDRESS for the scheduler cache.

--  6.16  2nd September 1991 (CEB)
--	  Added Atom_type by moving from Interrupt entries. This is the
--	  proper place for the definiton and also avoids pulling in too much
--	  of the interrupt rendezvous code if not used.
--	  Added Stack_elem_type for cleaner interrupt rendezvous code.

-- 6.17   3rd November 1991 (BJD)
--	  Prefix all CRTS library unit names with ALSYS_

-- 6.18   12th March 1992 (CEB)
--	  Added definitions of Interrupt Entry types.

-- 6.19   15th May 1992 (BJD)
--	  Complete the parameterizable types and constants for Interrupt
--	  Entries by adding MAX_HANDLERS and the INIT and ATTACH target types.

-- 6.20   19th May 1992 (BJD)
--	  Undo last change. The types belong in MACHINE_TASK_TYPES after all,
--	  since they are likely to involve a number of extra targ-dep type
--	  definitions, and maybe 'with' clauses, which do not belong in this
--	  spec.

-- 6.22   8th January 1993 (BJD)
--	  Add type definitions for the interface to unsigned enum rep values.

-- 6.25	  16th April 1993 (AJB)
--	  Re-target for HPPA.

-- 6.26   3rd June 1993 (AJB)
--	  Mod. special values of DATA_ADDRESS used by tasking kernel, to
--	  satisfy assertions in all cases (TARGET_HPPA).

-- 6.27   9th June 1993 (AJB)
--	  MASTER_LIST_TERMINATOR and WAIT_FOR_EVERYTHING must be possible stack
--	  values since OLDER comparisons on ADDRESSes must be in same space.

-- 6.29   7th October 1993 (BJD)
--        Add pragma ELABORATE for SYSTEM to avoid any chance of ABEs when
--        using SYSTEM.+ in any of the data initializations.

-- 6.31   26th October 1993 (Russell March)
--        Added operators for DATA_ADDRESS and definition of DATA_ADDRESS_SIZE 
--        for heap optimisation.

-- 6.33   18th February 1994 (BJD)
--        Correct syntax error, and tidy up comments and layout.
