-- $Source: /home/dsd/d/aa/monoBANK/AA/system.rts.tasking_interface.spc,v $ 
-- $Revision: 1.12 $ $Date: 1997/03/12 21:27:15 $ $Author: stt $ 
--        Copyright (C) 1995 by Intermetrics, Inc.   

with System.RTS.Interface_Types;      use System.RTS.Interface_Types; 
with System.RTS.Tasking_Types_Pkg;
with Ada.Exceptions;

package System.RTS.Tasking_Interface is

    pragma Suppress( Elaboration_Check );

--* OVERVIEW
-- This package contains interfaces the 9X code emitter for most
-- tasking operations.  The Handle types defined below are addresses of
-- data structures that are logically private to the RTS; the generated
-- code does not know the component structure. Their size is known to
-- the compiler, which generates code to allocate space for them. Generated
-- code must store handles received as out paramsters and functions 
-- return values from various RTS interfaces and use them
-- subsequently. For example, RTS_Add_Dependent takes a master handle as
-- context.

    subtype TCB_Handle is Simple_Handle;

    --[ 9.1 Task Types and Task Bodies

    procedure RTS_Initialize_Task(
        Number_Of_Entries : in Task_Entry_Index;  -- # entries in the task
        Act_List          : in Activation_Handle; -- Already initialized
        Master            : in Master_Handle;
        Code              : in Code_Address;      -- addr of task body code
        Static_Link       : in Data_Address;
        Task_Obj_Address  : in Data_Address;	  -- out parameter
        Priority          : in System.Priority'Base;
        Elab_Bit          : Tasking_Types_Pkg.Boolean_Ptr;
						-- Address of elab-bit
        Storage_Size      : in Integer
    );
    pragma export (Ada,
		   RTS_Initialize_Task,
                   "rts_initialize_task");
    pragma global (RTS_Initialize_Task);

    procedure RTS_Initialize_Task_With_Name(
        Number_Of_Entries : in Task_Entry_Index;  -- # entries in the task
        Act_List          : in Activation_Handle; -- Already initialized
        Master            : in Master_Handle;
        Code              : in Code_Address;      -- addr of task body code
        Static_Link       : in Data_Address;
        Task_Obj_Address  : in Data_Address;	  -- out parameter
        Priority          : in System.Priority'Base;
        Elab_Bit          : Tasking_Types_Pkg.Boolean_Ptr;
						-- Address of elab-bit
        Storage_Size      : in Integer;

	Name		  : in Data_Address       -- addr of nul-term'd string
						  -- or Null_Address
    );
    --* OVERVIEW
    -- Called for task creation.
    -- ..._With_Name version should be used to provide a name for the task
    --* EFFECTS
    -- Creates and initializes a TCB (Task Control Block) with the given
    -- properties, and returns a reference to the task in Task_Obj_Address.
    -- The created task is placed on the given activation list.
    -- The thread underlying the task is also created, in an unactivated
    -- state.
    pragma export (Ada,
		   RTS_Initialize_Task_With_Name,
                   "rts_initialize_task_with_name");
    pragma global (RTS_Initialize_Task_With_Name);
      
    --[ 9.2 Task Execution - Task Activation

    procedure RTS_Initialize_Activation_List(
	 List :  Activation_Handle  
    );
    --* OVERVIEW
    -- List contains a suspension object as well as a list of tasks
    -- to be activated.  List is created per object
    -- containing task components or declarative part, and is
    -- appended to in RTS_Initialize_Task.
    -- The suspension object provides a mechanism 
    -- for suspending while a group activation is in progress.
    -- The list must be protected from concurrent access
    -- after activation is started.
    --* EFFECTS
    -- List is initialized to an empty default state.
    
    pragma export (Ada,
		   RTS_Initialize_Activation_List,
                   "rts_initialize_activation_list");
    pragma global (RTS_Initialize_Activation_List);

    procedure RTS_Activate_Tasks(
	  List :  Activation_Handle
    );
    --* EFFECTS
    -- Activates all tasks on a list.
    pragma export (Ada, RTS_Activate_Tasks, "rts_activate_tasks");
    pragma global (RTS_Activate_Tasks);
    
    procedure RTS_Activation_Succeeded;
    --* EFFECTS
    -- Informs activator that the activation was successful.
    pragma export (Ada,
		   RTS_Activation_Succeeded,
                   "rts_activation_succeeded");
    pragma global (RTS_Activation_Succeeded);

    --[ 9.5.1 Entries and Accept Statements

    procedure RTS_Simple_Accept( 
        Which_Entry : in Task_Entry_Index;
        Params : out Param_Block_Ptr;
        Current_ECD : out ECD_Handle
    );
    --* OVERVIEW
    -- Called for simple accept statement. 
    --* EFFECTS
    -- Suspends until called task can start rendezvous.
    pragma export (Ada, RTS_Simple_Accept, "rts_simple_accept");
    pragma global (RTS_Simple_Accept);

    --[ 9.5.3 Entry Calls

    procedure RTS_Task_Entry_Call ( 
        Called_Task : Task_Obj_Handle; 
        Params : Param_Block_Ptr;
        Which_Entry : Task_Entry_Index
    );
    --* OVERVIEW
    -- Called for an unconditional task entry call.
    --* EFFECTS
    -- Allocates and initializes an ECD.  Calls a Task_Entry of
    -- Called_Task.Pr_TCB.
    pragma export (Ada,
		   RTS_Task_Entry_Call,
                   "rts_task_entry_call");
    pragma global (RTS_Task_Entry_Call);


    procedure RTS_Do_Task_Requeue(
        Current_ECD  : ECD_Handle;
        With_Abort   : Boolean;
    	Called_Task : Task_Obj_Handle; 
        Which_Entry  : Task_Entry_Index;
	Protected_To_Task : Boolean
    );
    --* OVERVIEW
    -- Called for a requeue statement to a task entry.
    pragma export (Ada,
		   RTS_Do_Task_Requeue,
                   "rts_do_task_requeue");
    pragma global (RTS_Do_Task_Requeue);

    --[ 9.7 Select Statements
    --[ The interfaces below build up a data structure that is 
    --[ kept in the TCB to describe a select statement.  

    procedure RTS_Start_Select(
        Header: Select_Handle; -- A Simple_Handle
        Number_Of_Alternatives : Accept_Alt_Index
    );
    --* OVERVIEW
    -- Called for a selective wait to describe the relationship
    -- between open accept alternatives and entries.  The map
    -- also includes the open alternative with minimum delay 
    -- and open terminate alternative (if any).
    --* EFFECTS
    -- Makes an empty from alternative indices to entry indices
    -- sized for this select. 
    pragma export (Ada,
		   RTS_Start_Select,
                   "rts_start_select");
    pragma global (RTS_Start_Select);

    procedure RTS_Add_Open_Accept(
        Which_Alt : Accept_Alt_Index;
        Which_Entry : Task_Entry_Index
    );
    --* OVERVIEW
    -- Called to add an open accept statement to the map.
    -- Assumed to be non-zero.
    pragma export (Ada,
		   RTS_Add_Open_Accept,
                   "rts_add_open_accept");
    pragma global (RTS_Add_Open_Accept);

    procedure RTS_Add_Open_Rel_Delay(
        Which_Alt : Accept_Alt_Index;
        Delay_Amount : Duration
    );
    --* OVERVIEW
    -- Called to add an open delay alternative to the map.
    --* EFFECTS
    -- Records only the minimum delay seen so far.
    -- Returns TRUE if this delay is minimum.
    pragma export (Ada,
		   RTS_Add_Open_Rel_Delay,
                   "rts_add_open_rel_delay");
    pragma global (RTS_Add_Open_Rel_Delay);

    procedure RTS_Add_Open_Abs_Delay(
        Which_Alt : Accept_Alt_Index;
        Expiration : Absolute_Delay_Handle;
        Which_Time_Type : Time_Type_Enum
    );
    --* OVERVIEW
    -- Called to add an open delay alternative to the map.
    --* EFFECTS
    -- Records only the minimum delay seen so far.
    -- Returns TRUE if this delay is minimum.
    pragma export (Ada,
		   RTS_Add_Open_Abs_Delay,
                   "rts_add_open_abs_delay");
    pragma global (RTS_Add_Open_Abs_Delay);

    procedure RTS_Mark_Terminate_Alt;
    pragma export (Ada,
		   RTS_Mark_Terminate_Alt,
                   "rts_mark_terminate_alt");
    pragma global (RTS_Mark_Terminate_Alt);

    procedure RTS_Mark_Else;
    pragma export (Ada,
		   RTS_Mark_Else,
                   "rts_mark_else");
    pragma global (RTS_Mark_Else);

    --[ 9.7.2 Conditional Entry Calls

--    procedure Do_Conditional_Entry_Call ( 

    function RTS_Conditional_Task_Call (
        Called_Task: Task_Obj_Handle;
        Params     : Param_Block_Ptr;
        Which_Entry : Task_Entry_Index
    ) return Boolean;
    --* OVERVIEW
    -- Called for a conditional task entry call.  Returns True if
    -- call completes immediately.
    pragma export (Ada,
		   RTS_Conditional_Task_Call,
                   "rts_conditional_task_call");
    pragma global (RTS_Conditional_Task_Call);
      
    procedure RTS_Do_Select(
        Alternative_Selected : out Accept_Alt_Index;
        Params : out Param_Block_Ptr;
        Current_ECD : out ECD_Handle
    );
    --* EFFECTS
    -- Wait for caller, group termination, or expiration
    -- depending on structure of select statement.
    -- The alternative index is returned when a caller
    -- becomes available before termination or delay expiration.
    pragma export (Ada,
		   RTS_Do_Select,
                   "rts_do_select");
    pragma global (RTS_Do_Select);



    procedure RTS_End_Of_Rendezvous;
    --* OVERVIEW
    -- Called at the end of the handled seq of statements
    -- inside an accept.
    pragma export (Ada,
		   RTS_End_Of_Rendezvous,
                   "rts_end_of_rendezvous");
    pragma global (RTS_End_Of_Rendezvous);
    
    procedure RTS_Except_End_Of_Rendezvous (
	Exc : Ada.Exceptions.Exception_Occurrence
    );
    -- Called at the end of the handled seq of statements
    -- inside an accept when an exception has been raised.
    pragma export (Ada,
		   RTS_Except_End_Of_Rendezvous,
                   "rts_except_end_of_rendezvous");
    pragma global (RTS_Except_End_Of_Rendezvous);


    --[ 9.7.3 Timed Entry Calls
    --  procedure Do_Timed_Entry_Call ( 
    
    function RTS_Timed_Rel_Task_Call (
        Called_Task  : Task_Obj_Handle; 
        Params       : Param_Block_Ptr;
        Which_Entry   : Task_Entry_Index;
        Delay_Amount : Duration
    ) return Boolean;
    --* OVERVIEW
    -- Called for a timed entry call with delay_relative.
    pragma export (Ada,
		   RTS_Timed_Rel_Task_Call,
                   "rts_timed_rel_task_call");
    pragma global (RTS_Timed_Rel_Task_Call);
      
    function RTS_Timed_Abs_Task_Call (
        Called_Task : Task_Obj_Handle;
        Params       : Param_Block_Ptr;
        Which_Entry  : Task_Entry_Index;
        Expiration  : Absolute_Delay_Handle;
        Which_Time_Type : Time_Type_Enum
    ) return Boolean;
    --* OVERVIEW
    -- Called for a timed entry call with delay_until.
    pragma export (Ada,
		   RTS_Timed_Abs_Task_Call,
                   "rts_timed_abs_task_call");
    pragma global (RTS_Timed_Abs_Task_Call );
      
    --[ 9.7.4 Asynchronous Transfer of COntrol

    function RTS_Async_Task_Call (
        ECD : ECD_Handle;
        Setjmp_Buffer : Data_Address;
        Master     : Master_Handle;
        Called_Task: Task_Obj_Handle;
        Params     : Param_Block_Ptr;
        Which_Entry : Task_Entry_Index
    ) return Boolean;
    --* OVERVIEW
    --  Called for an asynchronous entry call.  Returns True if the call
    -- completes immediately.
    pragma export (Ada,
		   RTS_Async_Task_Call,
                   "rts_async_task_call");
    pragma global (RTS_Async_Task_Call);
      

    --[ 9.9 Task and Entry Attributes

    function RTS_Is_Callable(
        Task_To_Check : Task_Obj_Handle
    ) return Boolean;
    --* OVERVIEW
    -- This corresponds to the attribute T'CALLABLE.
    --* EFFECTS
    -- Returns TRUE if the task is callable; FALSE otherwise.
    pragma export (Ada, RTS_Is_Callable, "rts_is_callable");
    pragma global (RTS_Is_Callable);

    function RTS_Task_Entry_Count(
        Which_Entry : in Task_Entry_Index
    ) return Caller_Count;                                   
    --* OVERVIEW
    -- This corresponds to the attribute E'COUNT.
    --* EFFECTS
    -- Returns the number of callers on a particular task entry.
    pragma export (Ada, RTS_Task_Entry_Count, 
			"rts_task_entry_count");
    pragma global (RTS_Task_Entry_Count);

    function RTS_Task_Entry_Caller(
        ECD : ECD_Handle
    ) return Task_Obj_Handle;
    --* OVERVIEW
    -- This corresponds to the attribute E'Caller.
    --* EFFECTS
    -- Returns the task ID of current caller of a particular task entry.
    pragma export (Ada, RTS_Task_Entry_Caller, 
			"rts_task_entry_caller");
    pragma global (RTS_Task_Entry_Caller);

    
    function RTS_Is_Terminated( 
    	Task_To_Check : Task_Obj_Handle 
    ) return Boolean;
    --* OVERVIEW
    -- Supports attribute T'Terminated.
    --* EFFECTS
    -- Returns TRUE if the task is terminated; false otherwise.
    -- This corresponds to the attribute T'TERMINATED.
    pragma export (Ada, RTS_Is_Terminated, "rts_is_terminated");
    pragma global (RTS_Is_Terminated);

    function RTS_Task_Storage_Size(
        Task_To_Check: Task_Obj_Handle    -- Task of interest
    ) return Integer;
    --* EFFECTS
    -- Returns the size of the task's (primary) stack in storage
    -- units. This corresponds to the attribute T'STORAGE_SIZE.
    pragma export (Ada,
		   RTS_Task_Storage_Size,
                   "rts_task_storage_size");
    pragma global (RTS_Task_Storage_Size);




    --[ 9.8 Abort of a Task

    procedure RTS_Abort_Task(
        Aborting : in Task_Obj_Handle
    );
    --* EFFECTS
    -- Aborts a task. This corresponds to an Ada ABORT statement.
    pragma export (Ada, RTS_Abort_Task, "rts_abort_task");
    pragma global (RTS_Abort_Task);

    
end System.RTS.Tasking_Interface;




