-- $Source: /home/dsd/d/aa/monoBANK/AA/system.rts.tasking_types_pkg.spc,v $
-- $Revision: 1.34 $ $Date: 1997/02/26 15:47:32 $ $Author: bethc $
--        Copyright (C) 1995 by Intermetrics, Inc.   

with Ada.Exceptions;
with System.RTS.Exception_Types;
with System.RTS.Interface_Types;
with System.RTS.Protected_Types_Pkg;
with System.RTS.TGT.Frames;
with System.RTS.TGT.Kernel.Threads;
with System.RTS.Timer_Class;
with Ada.Finalization;

package System.RTS.Tasking_Types_Pkg is

    pragma Suppress( Elaboration_Check );

--* OVERVIEW
-- This package contains the Task Control Block (TCB) and component types.
-- The activation record and master record are also defined here because
-- of the interdependencies between TCBs and there data structures.
--
-- A TCB contains master components: (1) the enclosing master and (2) the
-- "innermost master" that the task is executing.  A master record
-- contains a pointer to the TCB that is executing it.

-- A TCB contains a pointer to the activation record for the
-- group activation that is currently in progress (if any).
-- This component is need so that the a newly-activated task can
-- notify its master.  The activation record contains a list of
-- tasks to be activated.

    package PRT renames Protected_Types_Pkg;
    package INT renames Interface_Types;

    Tasking_Internal_Error : exception;
	
    subtype ECD_Access is PRT.ECD_Access;
    subtype Task_Entry_Index is INT.Task_Entry_Index;
    subtype Timer_Class_Container is Timer_Class.Timer_Class_Container;

    type TCB_Type( Num_Entries : Task_Entry_Index);
    type TCB_Access is access all TCB_Type;
    type Boolean_Ptr is access all Boolean;

    -- Generation number is kept for dangling pointer check.
    -- Generated code stores values of TCB_Handle type. When a task
    -- terminates, it's storage is deallocated.  To detect the use of
    -- a handle for a terminated task, the generation numbers in the
    -- encloser and TCB proper are compared.

    type TCB_Gen_Num is new Natural;

    -- First two values of TCB_Gen_Num used for null and environment task.
    NOT_IN_USE             : constant TCB_Gen_Num  := TCB_Gen_Num'First;
    TCB_Gen_Count          : TCB_Gen_Num           := TCB_Gen_Num'First + 1;
    				-- Incremented before use

    type Task_Obj_Header is record
        TCB     : TCB_Access;
        Gen_Num : TCB_Gen_Num := NOT_IN_USE;
    end record;

    type Task_Obj_Access is access all Task_Obj_Header;
    -- Note: We are depending on static initialization of the following
    -- special object headers
    Null_Task_Obj_Header : constant
    	Task_Obj_Header := Task_Obj_Header'( TCB => null,
	    	    	                     Gen_Num => TCB_Gen_Num'First);

    -- Task lists are used for activations, dependencies and free
    -- TCBs.  Links are defined in TCB.  Only one pointer is needed
    -- in the head since none of these lists must be sorted.

    type Task_List_Head is record
       Count : Natural := 0;
       First : TCB_Access;
    end record;

    subtype Task_List_Type is Task_List_Head;

      -- Note that Runnable may be the status even after the task has
      -- been aborted or has completed, because the task is
      -- performing cleanup activities.

    type Task_Status is (
      	Unactivated,      -- Created & elaborated; on activation list.
      	Runnable,         -- Not suspended in kernel. Distinction between
        		  -- runnable and running is made in kernel.
      	--
      	-- The remaining states (except Unknown) are a breakdown of the
      	-- Suspended State in the kernel.
      	--
      	Task_Wait,              -- Waiting for dependents to terminate.
      	Terminated,             -- Terminated; all dependents terminated;
                                -- no longer running stack space released.
      	Delayed,                -- Waiting for delay to expire; on delay queue.
      	Activator_Waiting,      -- Waiting for activation of subtasks;
      	Caller_Waiting,         -- Normal entry call; on entry queue for
                                -- either protected object or task.
      	Caller_Delayed,         -- Timed entry call; on delay and entry queue;
      	Acceptor_Waiting,       -- At ACCEPT, awaiting a caller; entries open.
      	Acceptor_Delayed,       -- SELECTIVE WAIT with delay; on delay
      	      	    	        -- queue; entries open.
        Acceptor_Terminable     -- SELECTIVE WAIT with terminate; entries open.
    );


    -- Completion Status

    type Completion_Status_Enum is (
      	Not_Yet_Completed,       -- Initial value
      	Completed_Normally,      -- Normal completion of task body
      	Completed_In_Select,     -- Completion due to selecting
	    	    	    	 -- terminate alternative.
      	Never_Activated,         -- Task terminated without activating.
      	Unhandled_Exception,     -- Completion due to unhandled exception.
      	Aborted                  -- Completed due to abort of task
	    	    	    	 -- or ancestor.
    );

    subtype Alternative_Index is INT.Accept_Alt_Index;
      -- index of accept alternative in a select statement

    type Entry_For_Alt is array( Alternative_Index range <> )
      of Task_Entry_Index;
      -- Array that indicates which entry is accepted by each accept
      -- alternative; zero means accept alternative is closed


    type Select_Record( Last_Alt : Alternative_Index );
    type Select_Access is access all Select_Record;

    type Select_Record( Last_Alt : Alternative_Index ) is record
        Has_Terminate   : Boolean;
        Has_Else        : Boolean;
        Delay_Alt_Index : Alternative_Index'Base;
        Rel_Delay_Value : Duration;
        Time_Type       : INT.Time_Type_Enum;
        Abs_Delay_Value : aliased Timer_Class_Container;
        -- Indicates accept alternative for best caller
        Alt_Index_Chosen : Alternative_Index'Base;
	Some_Alternative_Open : Boolean; -- accept or delay
        Entry_List      : Entry_For_Alt( 0 .. Last_Alt );
 	-- Entry_List is last field so we do not allocate an extra
	-- word for a pointer to the dynamic part.
    end record;

    Max_Accept_Nesting : constant := 12;  -- ???

    subtype Accept_Nesting_Level is Natural range 0..Max_Accept_Nesting;    

    type Pr_TCB_Type;
	
    type Pr_TCB_Type_Access is access Pr_TCB_Type; 
      -- needed as target of type cast from PO_Access.

    -- Only instantiaions of Task_Attribute_ID have visibility to the
    -- attribute type.
    type Dummy_Attribute_Value is record null; end record;

    type Attribute_Handle is access all Dummy_Attribute_Value;

    type Attribute_ID is new Positive;

    Last_Attribute_ID_Preallocated : constant := 4;

    Last_Attribute_ID : constant := 32;

    type Attribute_Index_Preallocated is array(Attribute_ID
        range 1..Last_Attribute_ID_Preallocated) of
        Attribute_Handle;

    type Attribute_Index_Extension is array(Attribute_ID range <>) of
        Attribute_Handle;

    type Attribute_Index_Extension_Access is access Attribute_Index_Extension;

    type Attribute_Cleanup_Op is access protected procedure(
	Handle : Attribute_Handle);

    -- Each instance of Ada.Task_Attributes declares an object of this type.
    -- The object finalization is really the finalization of the instance.
    -- All current attributes residing in TCB's are deallocated and, hence, 
    -- finalized.
    type System_Wide_Attribute_Cleanup_Type is new Ada.Finalization.Controlled 
        with record 
	    ID : Attribute_ID; 
        end record;

    protected type Pr_TCB_Type (
	Num_Entries : Task_Entry_Index
    ) is
    --* OVERVIEW
    -- This type is the "protected" part of the TCB.
    -- Components are included here if they may be
    -- accessed by more than one task.  Also, protected
    -- operations provide the basis for rendezvous.
	
        pragma Interrupt_Priority( Any_Priority'Last );

        procedure Default_Reinitialize(Enclosing_TCB : TCB_Access);
        --* OVERVIEW
        -- Reinitializes to same values as is done for a new object,
	-- allowing re-use.

        entry Task_Entry(1..Num_Entries)(
            Params : PRT.Param_Block_Ptr
        );
        --* OVERVIEW
        -- A task entry call becomes a call on a member
        -- of this family.
	
       	entry Simple_Accept(
	    Select_Info : Select_Access;  
	    Alt_Index_Chosen : out Alternative_Index;  
	    Params : out PRT.Param_Block_Ptr;
	    ECD : out PRT.ECD_Access
        );
	--* OVERVIEW
        -- Called for an accept statement not nested in selective
        -- accept.

        entry Selective_Accept(
	    Select_Info : Select_Access;  
            Alt_Index_Chosen : out Alternative_Index;  
            Params : out PRT.Param_Block_Ptr;
            Call_Info : out PRT.ECD_Access
        );
        --* OVERVIEW
        -- A selective accept becomes a call on this entry.
        -- If the selective accept has an open delay alternative,
        -- then a timed entry call on this entry is used.
        -- If the selective accept has an open alternative a simple
        -- call on this entry is used.
        -- The call is followed by a "case" statement to execute
        -- the appropriate accept alternative.
        --* EFFECTS
        -- If there is a caller, then the highest is chosen.  The
        -- accept alternative corresponding to this caller is
        -- returned. If there is no caller, requeue on a private
        -- entry until some caller becomes available.
        -- Actual parameters are obtained from ECD.
        -- Notify master that this task is terminable if
        -- no callers and Terminable = True.

        function Is_Callable return Boolean;
        --* OVERVIEW
        -- Function to read Callable attribute.

        procedure Set_Uncallable;
        --* OVERVIEW
        -- Sets Callable attribute to False;

        function Get_Base_Priority return System.Any_Priority;
        --* OVERVIEW
        -- Function to read Base_Priority attribute.

        function Get_Active_Priority return System.Any_Priority'Base;
        --* OVERVIEW
        -- Function to read Base_Priority attribute.

        procedure Set_Base_Priority(Priority : System.Any_Priority);
        --* OVERVIEW
        -- Sets Base_Priorty attribute.

        procedure Inherit_Priority(Priority : System.Any_Priority);
        --* OVERVIEW
        -- Sets Active_Priority to the max of inherited priorities.
	-- Note: Excepting priority inheritance during protected
	-- actions, we may only have a single source of inheritance.

        procedure Disinherit_Priority(Priority : System.Any_Priority);
        --* OVERVIEW
        -- Sets Active_Priority to Base_Priority.

        function Get_Aborting_Level return Natural;
        --* OVERVIEW
        -- Returns the outermost level to which the task is
        -- aborting.

        procedure Reset_Aborting_Level;
        --* OVERVIEW
        -- Resets the aborting level on a successfull abort.

        function Get_Aborting_ECD return ECD_Access;
        --* OVERVIEW
        -- Returns the ECD of the outermost level to which the task is
        -- aborting.  Null if an abort is not in progress.
	
        procedure Abort_Task;
        --* OVERVIEW
        -- Procedure to abort a task.
	
        procedure Abort_To_Level(
            ECD : ECD_Access
        );
        --* OVERVIEW
        -- Procedure to abort an abortable part.  If level in ECD
        -- is lower than current abort in progress, if any, then
        -- this task is sent an abort signal.

         procedure End_Of_Rendezvous ( ECD : out PRT.ECD_Access );
         --* OVERVIEW
         -- Called at end of rendezvous to release the caller.

	 procedure Except_End_Of_Rendezvous(
	    Exc : Ada.Exceptions.Exception_Occurrence
	);
         --* OVERVIEW
         -- Called at end of rendezvous to release the caller.
         -- Exc is exception raised during rendezvous

	procedure Hold;
	--* OVERVIEW
	--  Called by Ada.Dynamic_Priorities.Hold.

	procedure Continue;
        --* OVERVIEW
	-- Called by Ada.Dynamic_Priorities.Continue.

	function Is_Held return Boolean;
	--* OVERVIEW
	--  Called by Ada.Dynamic_Priorities.Is_Held.

	procedure Recompute_Priorities;
	--* Called when an inherited priority has changed.

        -- operations for Ada.Task_Attributes;
        function Value(ID : Attribute_ID) return Attribute_Handle;

        function Reference(ID : Attribute_ID) return Attribute_Handle;

        procedure Set_Value(
            Val : in Attribute_Handle;
            ID : Attribute_ID;
            Old_Val : out Attribute_Handle
        );

	procedure Finalize_Attribute(
	    ID : Attribute_ID
	);

    private

        entry Caller_In_Rendezvous( Accept_Nesting_Level'range) (
            Params : PRT.Param_Block_Ptr
	);
        --* OVERVIEW
        -- Caller requeued here during rendezvous
	
        entry Wait_For_Caller(Accept_Nesting_Level'range) (
	    Select_Info : Select_Access;  
	    Alt_Index_Chosen : out Alternative_Index;  
	    Params : out PRT.Param_Block_Ptr;
	    Call_Info : out PRT.ECD_Access
        );
        --* OVERVIEW
	-- Acceptor requeued here if no caller

	function Is_Open(
	    Ent : Task_Entry_Index
	) return Boolean;
        --* OVERVIEW
	-- Called in barrier of Task_Entry.

        -- Protected task state information follows.
	-- Note: explicit default initialization should match the
	-- re-initialization in Default_Reinitialize.

	-- points to encloser, corresponds to thread ptr. in kernel
	Enclosing_TCB : TCB_Access;

        -- Non-null while caller in rendezvous
        Current_Nesting_Level : Accept_Nesting_Level := 0;
        Current_Call_Info : ECD_Access;

        Current_Params : PRT.Param_Block_Ptr;
	Current_Select_Info : Select_Access;

        -- 'Callable
        Callable : Boolean := True;

        -- At an open terminate alternative
        -- Next two fields are protected because multiple triggers
        -- are possible with nested asynchronous selects.
        Aborting_Level : Natural := Natural'Last;
        Aborting_ECD : ECD_Access;

        -- Must include priority because self and other thread
        -- could be setting priority at same time, though this
        -- is highly unlikely.
        -- kernel doesn't keep this
        Base_Priority : Any_Priority := Any_Priority'First;

        Contributing_Base_Priority : Any_Priority'Base := Any_Priority'First;
	-- base priority that is considered for priority inheritance
	-- same as Base_Priority unless task is held (ATC.Hold).

        Active_Priority : Any_Priority'Base := Any_Priority'First;

        Attribute : Attribute_Index_Preallocated;

        Attribute_Extension : Attribute_Index_Extension_Access;

    end Pr_TCB_Type;


    -- The object below is used to save the type descriptor generated
    -- for Pr_TCB_Type.  It is initialized in RTS_Init and
    -- used to test for special cases involving a Pr_TCB protected object.

    TCB_Pr_Type_Desc_Access : PRT.Pr_Type_Desc_Access;

    type Master_Record;
    type Master_Access is access all Master_Record;
	
    subtype Frame_ID_Type is TGT.Frames.Frame_ID;

    subtype Dependency_List is Task_List_Type;

    type Cleanup_Record;

    type Cleanup_Ptr is access all Cleanup_Record'Class;
    for Cleanup_Ptr'Storage_Size use 0;
        -- No allocators, and no need to register collection for finalization

    type Cleanup_Record is abstract tagged record 
	Previous : Cleanup_Ptr;
    end record;

    procedure Cleanup(Rec : Cleanup_Record) is abstract;

    type Master_Record is record
	Dependents          : Dependency_List;
          -- dependent tasks created by a task declaration
          -- or allocated with allocator type in this master
        Nonterminable_Count : Natural := 1;
          -- used to decide when to accept terminate alternative
          -- and to wait for dependent tasks
	Cleanup_Actions     : Cleanup_Ptr;
          -- actions to cleanup protected objects or objects with user-defined
	  -- finalization
        Task_Body_Master    : Boolean := False;
          -- if this master is a task body master, the task must
          -- be marked uncallable upon reaching finalization
	Enclosing_Master    : Master_Access;
          -- dynamically enclosing master
	Enclosing_Task      : TCB_Access;
          -- task executing this master
	Trigger_ECD         : ECD_Access;
          -- the ECD for the trigger in an async select
	Dependents_Complete : TGT.Kernel.Threads.Suspension_Object;
          -- used for awaiting dependent tasks
	Frame_ID            : Frame_ID_Type := TGT.Frames.Null_Frame_ID;
          -- used to identify subprogram master for exception propagation
	Finalize_Errors_To  : Address;
	  -- address at which Program_Error should be raised if an
	  -- exception is raised during finalization of objects associated
	  -- with this master
    end record;		-- Default initializers used for Default_Master
	
    Default_Master : aliased Master_Record; 

    -- The activation record is used to synchronize activity of
    -- activator and the group of tasks that it is currently activating.
    -- A component of this type is kept in the TCB so that activating
    -- tasks can find it for notification.

    type Activation_List;
    type Activation_Access is access all Activation_List;

    type Activation_List is record
        To_Be_Activated        : Task_List_Type;
        Some_Activation_Failed : Boolean := False;
	pragma Atomic(Some_Activation_Failed);
	Activations_Complete   : TGT.Kernel.Threads.Suspension_Object;
    end record;

    type TCB_Type( Num_Entries : Task_Entry_Index) is new
	    TGT.Kernel.Threads.Thread_Descriptor with record
	-----------------------------------------------------------------
        --- Task Identification ---
	-----------------------------------------------------------------

	Task_Obj	: Task_Obj_Access; -- ptr to task object
	Gen_Num		: TCB_Gen_Num;	  -- Copy for validity check 
        Static_Link	: INT.Data_Address;
        Code		: INT.Code_Address;
	Storage_Size	: Integer;
        Elab_Bit	: Boolean_Ptr;
	-----------------------------------------------------------------
        --- State Information ---
	-----------------------------------------------------------------
        Status		: Task_Status := Unactivated;
	    -- put Status in Pr_TCB_Pkg?
        Completion_Status : Completion_Status_Enum := Not_Yet_Completed;

	Occurrence	: Exception_Types.Minimal_Exception_Occurrence;
    	Blocking_ECD	: ECD_Access;
        Activation_Sync	: Activation_Access;
          -- assoc. w/ Activator_Waiting
	Async_Level	: Natural := 0;  
	  -- nesting level of asynchronous selects

	-----------------------------------------------------------------

        --- Scope Information ---
	-----------------------------------------------------------------
        Master		: Master_Access; -- of this task.
        Innermost_Submaster : Master_Access; -- master most
	    	    	    	     	    	 -- recently entered
        Activator	: TCB_Access;     -- task that act'd this one
	-----------------------------------------------------------------
        --- Select Info ---
	-----------------------------------------------------------------
        Select_Info	: Select_Access;
          -- really need a discriminant for # accept alternatives
	-----------------------------------------------------------------
        --- Links ---
	-----------------------------------------------------------------
        Act_Prev	: TCB_Access;     -- activation list
        Act_Next	: TCB_Access;
        Dep_Prev	: TCB_Access;     -- dependency list
	Dep_Next	: TCB_Access;
	-----------------------------------------------------------------
        --- Entry Information ---
	-----------------------------------------------------------------
        -- put at end for ease of access at initialization time
        Pr_TCB	: Pr_TCB_Type(Num_Entries => Num_Entries);
    end record;

    procedure Release_Thread(TCB : access TCB_Type);
	-- This operation is called by Finish_Thread to allow reuse of
	-- a TCB that has been placed in the free list by Free_TCB.

    function Self return TCB_Access;
    --* Returns pointer to TCB of the task that is executing.
    pragma Inline(Self);	-- Allow special treatment of current_thread

    function New_TCB return TCB_Access;
    --* Returns pointer to a new or reused available TCB.

    procedure Free_TCB(TCB : TCB_Access);
    --* Returns TCB to free list marked as still in use.

    procedure Default_Reinitialize_TCB (TCB : TCB_Access);
    --* Needed for init. of envr. TCB

    function Assign_Attribute_ID(
        Cleanup_Op : Attribute_Cleanup_Op
    ) return Attribute_ID;
    --* Registers attribute Cleanup routine in system-wide table and returns
    -- next available ID. 

    procedure Deassign_Attribute_ID(
        ID : Attribute_ID
    );
    --* Makes attribute ID reusable. 

    procedure Finalize(Obj : in out System_Wide_Attribute_Cleanup_Type);

end System.RTS.Tasking_Types_Pkg;

