-- $Source: /home/dsd/d/aa/monoBANK/AA/TSP_system.rts.tgt.sec_stack_pkg.bdy,v $ 
-- $Revision: 1.5 $ $Date: 1996/11/19 12:26:50 $ $Author: akira $ 

with System;
with System.RTS.Heap_Storage_Pkg;
with System.Storage_Elements;
with SYSTEM.RTS.TCB;
with System.RTS.Traces;
with Unchecked_Conversion;

package body System.RTS.TGT.Sec_Stack_Pkg is

    pragma Suppress (All_Checks);
    
    type Mark_Object is
       record
          Saved_Sec_Stack_Ptr : System.Address;
       end record;

    Bad_SS_Mark : exception;


    --* OVERVIEW
    -- Space on the secondary stack should be managed as a mark-release heap.
    -- System.RTS.Secondary_Stack_Pkg should be implemented in such a way
    -- that one can obtain the highwater mark of the secondary stack
    -- utilization (Mark), allocate a given number of storage units of memory
    -- on the secondary stack (Allocate), and release memory so that 
    -- the secondary stack is restored back to a given mark (Release).

    --* IMPLMENTATION
    -- The secondary stack is implemented as a chain of allocated chunks
    -- of memory. Each chunk points to the previous chunk (the one allocated
    -- immediately prior to this chunk) or null if it is the first chunk.
    -- Each chunk also has a pointer to the last address in the chunk. An
    -- address is determined to be in the range of a chunk by scanning the
    -- chain until a chunk is found with an address range that contains the
    -- particular address.
    -- The TCB contains a pointer to the last allocated chunk and the
    -- highwater mark for the stack.  There is always at least one chunk
    -- on the list.  The highwater mark is an address in the range of a chunk.


    -- These routines are used by the general heap package, too.
    function rts_heapalloc (dwbytes : in Storage_Elements.Storage_Count) return Address;
    pragma Import (C, rts_heapalloc, "rts_heapalloc");

    procedure rts_heapfree (addr : in address);
    pragma Import (C, rts_heapfree, "rts_heapfree");
    -- Note that we do not set the pointer to null afterwards; it is 
    -- not needed in this context.

  
    Allocation_Alignment : constant := 8;
    
    type Chunk_Header;  
    type Chunk_Ptr is access all Chunk_Header;
    
    type Chunk_Header is 
       record
          Previous  : Chunk_Ptr;
          Chunk_End : System.Address;
       end record;

    for Chunk_Header use
       record
          Previous  at 0 range 0..31;
          Chunk_End at 4 range 0..31;
       end record;
    Chunk_Header_Size : constant := 8;
       -- (Chunk_Header'Size/System.Storage_Unit - 1) / Allocation_Alignment + 1;


    Minimum_Chunk_Data_Size : constant := 4096 - Chunk_Header_Size;
    -- The size of the user data in the chunk.
    

    function To_Chunk_Ptr is 
       new Unchecked_Conversion (System.Address, Chunk_Ptr);
    function To_Address is
       new Unchecked_Conversion (Chunk_Ptr, System.Address);
       
    
    function Align_Extend (Size : in Integer) return Natural;
    pragma Inline (Align_Extend);

    function Align_Extend (Size : in Integer) return Natural is
       -- Extend Size to the next multiple of alignment.
    begin
       return Size + ((-Size) mod Allocation_Alignment);
    end Align_Extend;


    function Max (Left, Right : in Natural) return Natural;
    pragma Inline (Max);

    function Max (Left, Right : in Natural) return Natural is
       -- Return the larger of the parameters
    begin
       if Left >= Right then
          return Left;
       else
          return Right;
       end if;
    end Max;
    

    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --


    procedure Mark( Mark_Save : Mark_Ptr ) is
        --* OVERVIEW
        -- Mark allows an application to "remember" or "mark"
	-- its current point of secondary stack utilization.
        --
        --* EFFECTS
        -- Mark will save the current 'Mark' of secondary stack
        -- utilization in the area addressed by Mark_Save.
        -- A subsequent call to Release with this same mark will
        -- nominally restore the secondary stack to the same point,
        -- releasing any secondary stack allocated in the interim.
    begin
       Mark_Save.Saved_Sec_Stack_Ptr := SYSTEM.RTS.TCB.Curr_TCB.Highwater_Mark;
    end Mark;


    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --


    procedure Initialize (Size : in Natural; 
		Current_TCB : System.RTS.TCB.Ref_TCB := Null) is
       -- Allocate and initialize the initial chunk for a secondary stack.
       -- Initializes the SS components of current TCB for this new chunk.  
       -- This function should be called for each created thread to initialize
       --  its secondary stack.

       -- Allocate and initialize a new chunk
       TCB_Ptr : SYSTEM.RTS.TCB.Ref_TCB := Current_TCB;
       New_Chunk_Adrs : System.Address := rts_heapalloc (
                  System.Storage_Elements.Storage_Count (Size));
       use type System.RTS.TCB.Ref_TCB;
    begin

       -- Use the current TCB is none was specified
       if TCB_Ptr = Null then
          TCB_Ptr := System.RTS.TCB.Curr_TCB;
       end if;

       if New_Chunk_Adrs = Null_Address then
          TCB_Ptr.Last_Chunk     := null;
          TCB_Ptr.Highwater_Mark := null;
          raise Storage_Error;
       end if;

       To_Chunk_Ptr (New_Chunk_Adrs).Previous  := null;
       To_Chunk_Ptr (New_Chunk_Adrs).Chunk_End := New_Chunk_Adrs + Size;
    
       -- Update the TCB
       TCB_Ptr.Last_Chunk     := New_Chunk_Adrs;
       TCB_Ptr.Highwater_Mark := New_Chunk_Adrs + Chunk_Header_Size;

    end Initialize;



    function Allocate( Size : in Natural ) return System.Address is
        --* OVERVIEW
        -- Allocate provides the basic mechanism of secondary stack
        -- memory allocation.
        --
        --* EFFECTS
        -- Allocate will reserve a block of space equal to the size
        -- requested and aligned to the proper target dependent boundary
	-- and return its address.  If sufficient space is not available,
        -- STORAGE_ERROR will be raised and the state of the secondary
        -- stack remains unchanged.
        --
	-- Before Allocate:
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            |
        --        |        |            |
        --        |        |------------| Highwater mark (M)
        --   direction     |            |
        --      of         |            |
        --     stack       |            |
        --     growth      |            |
        --        |        |            |
        --        |        |            |
        --        V        |            |
        --                 |            |
        --                 |            |
        --                 |            |
        --                       .
        --                       .
        --                       .
	--
	--      New_Stack := Allocate( Size => Needed_Stack );
	--
	-- After Allocate:
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            |
        --        |        |            |
        --        |        |------------| <-- New_Stack
        --   direction     |            |
        --      of         | Newly      |
        --     stack       |  Allocated |
        --     growth      |   Stack    |
        --        |        |    Space   |
        --        |        |            |
        --        V        |------------| Highwater mark 
        --                 |            | (M+Needed_Stack+Alignment)
        --                 |            |
        --                 |            |
        --                       .
        --                       .
        --                       .
	--
        -- The memory address after proper alignment of the allocated
	-- secondary stack space is returned in New_Stack and the 
	-- internal highwater mark set to (M+Needed_Stack+Alignment)
	-- storage units.

       Curr_TCB        : SYSTEM.RTS.TCB.Ref_TCB := SYSTEM.RTS.TCB.Curr_TCB;       
       Last_Chunk      : Chunk_Ptr := To_Chunk_Ptr (Curr_TCB.Last_Chunk);
       HW_Mark         : System.Address := Curr_TCB.Highwater_Mark;
       Allocation_Size : Natural := Align_Extend (Size); 
       Chunk_Size      : Natural;
       
    begin
       
       -- Will the allocation fit in current chunk?  If it will, then
       -- bump adjust highwater mark.  If not, then allocate another
       -- chunk, and then adjust the highwater mark.
       if Last_Chunk/=null and then -- sdf
          (Last_Chunk.Chunk_End - HW_Mark) >= Allocation_Size then
          Curr_TCB.Highwater_Mark := HW_Mark + Allocation_Size;
          return HW_Mark;
          
       else
          -- Allocate and initialize a new chunk
          Chunk_Size := Max (Allocation_Size, Minimum_Chunk_Data_Size) + 
                        Chunk_Header_Size;
          declare
             New_Chunk_Adrs : System.Address := rts_heapalloc (
                        System.Storage_Elements.Storage_Count (Chunk_Size));
          begin
             if New_Chunk_Adrs = Null_Address then
                raise Storage_Error;
             end if;
             To_Chunk_Ptr (New_Chunk_Adrs).Previous  := Last_Chunk;
             To_Chunk_Ptr (New_Chunk_Adrs).Chunk_End := New_Chunk_Adrs +
                                                        Chunk_Size;
          
             -- Update the TCB
             Curr_TCB.Last_Chunk := New_Chunk_Adrs;
             Curr_TCB.Highwater_Mark := 
                New_Chunk_Adrs + Chunk_Header_Size + Allocation_Size;
              
             return New_Chunk_Adrs + Chunk_Header_Size;
          end;
       end if;

    end Allocate;


    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --


    function Allocate_Without_Exceptions ( Size : in Natural ) return System.Address is

        --* OVERVIEW
        -- Behaves like Allocate except that in the case where Allocate
        -- raises Storage_Error, this routine returns a System.Null_Address;

       Curr_TCB        : SYSTEM.RTS.TCB.Ref_TCB := SYSTEM.RTS.TCB.Curr_TCB;       
       Last_Chunk      : Chunk_Ptr := To_Chunk_Ptr (Curr_TCB.Last_Chunk);
       HW_Mark         : System.Address := Curr_TCB.Highwater_Mark;
       Allocation_Size : Natural := Align_Extend (Size); 
       Chunk_Size      : Natural;
       
    begin
       
       -- Will the allocation fit in current chunk?  If it will, then
       -- bump adjust highwater mark.  If not, then allocate another
       -- chunk, and then adjust the highwater mark.
       if Last_Chunk/=null and then -- sdf
          (Last_Chunk.Chunk_End - HW_Mark) >= Allocation_Size then
          Curr_TCB.Highwater_Mark := HW_Mark + Allocation_Size;
          return HW_Mark;
          
       else
          -- Allocate and initialize a new chunk
          Chunk_Size := Max (Allocation_Size, Minimum_Chunk_Data_Size) + 
                        Chunk_Header_Size;
          declare
             New_Chunk_Adrs : System.Address := rts_heapalloc (
                        System.Storage_Elements.Storage_Count (Chunk_Size));
          begin
             if New_Chunk_Adrs = Null_Address then
                return Null_Address;
             end if;
             To_Chunk_Ptr (New_Chunk_Adrs).Previous  := Last_Chunk;
             To_Chunk_Ptr (New_Chunk_Adrs).Chunk_End := New_Chunk_Adrs +
                                                        Chunk_Size;
          
             -- Update the TCB
             Curr_TCB.Last_Chunk := New_Chunk_Adrs;
             Curr_TCB.Highwater_Mark := 
                New_Chunk_Adrs + Chunk_Header_Size + Allocation_Size;
              
             return New_Chunk_Adrs + Chunk_Header_Size;
          end;
       end if;

    end Allocate_Without_Exceptions;
       

    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --


    procedure Release( Mark_Save : in Mark_Ptr ) is

        --* OVERVIEW
        -- Release provides for the release of memory allocated
        -- from the secondary stack.
        --
        --* EFFECTS
	-- Memory allocated between the time the Mark_Save was established
	-- (via a call to Mark) and the call to Release will be released.
        -- Release will return the secondary stack to the condition it was
	-- at the point where the 'Mark' given by Mark_Save was established.
        --
        -- Before Release:
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            |
        --        |        |            |
        --        |        |            | (Previous_Mark)
        --   direction     |            |
        --      of         |            |
        --     stack       |            |
        --     growth      |            |
        --        |        |            |
        --        |        |            |
        --        V        |------------| Highwater mark (M)
        --                 |            |
        --                 |            |
        --                 |            |
        --                       .
        --                       .
        --                       .
	--
	--      Release( Mark_Save => Save_Address );
	--
	-- After Release:
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            |
        --        |        |            |
        --        |        |------------| Highwater mark (Previous_Mark)
        --   direction     |            |
        --      of         |            |
        --     stack       |            |
        --     growth      |            |
        --        |        |            |
        --        |        |            |
        --        V        |            | (M)
        --                 |            |
        --                 |            |
        --                 |            |
        --                       .
        --                       .
        --                       .
	--

       Curr_TCB   : SYSTEM.RTS.TCB.Ref_TCB := SYSTEM.RTS.TCB.Curr_TCB;       
       Last_Chunk : Chunk_Ptr := To_Chunk_Ptr (Curr_TCB.Last_Chunk);
       Temp_Chunk : Chunk_Ptr;

    begin
       
       --> Remove Debug code
       -- if System.RTS.Traces.TRACE (49) then -- ignore bad SS releases
       --   -- Check before deallocating any chunks.
       --   while Mark_Save.Saved_Sec_Stack_Ptr < To_Address (Last_Chunk) or else
       --         Mark_Save.Saved_Sec_Stack_Ptr > Last_Chunk.Chunk_End loop
       --      Temp_Chunk := Last_Chunk;
       --      Last_Chunk := Last_Chunk.Previous;
       -- 
       --       if Last_Chunk=null then
       --          return; -- ignore bad release
       --      end if;
       --    end loop;
       -- end if;

       -- Traverse the chunk chain until we find the chunk that contains the 
       -- mark, deallocating the chunks along the way.  Then, set the 
       -- highwater mark to the given mark.
       -- Typically the body of this loop will not be executed.
       while Mark_Save.Saved_Sec_Stack_Ptr < To_Address (Last_Chunk) or else
             Mark_Save.Saved_Sec_Stack_Ptr > Last_Chunk.Chunk_End loop
          Temp_Chunk := Last_Chunk;
          Last_Chunk := Last_Chunk.Previous;
          Curr_TCB.Last_Chunk := To_Address (Last_Chunk);
          rts_heapfree (To_Address (Temp_Chunk));

          if Last_Chunk=null then
             raise Bad_SS_Mark;
          end if;
       end loop;
       
       Curr_TCB.Highwater_Mark := Mark_Save.Saved_Sec_Stack_Ptr;

    --> Remove Debug code
    -- exception
    --    when others =>
    --       if System.RTS.Traces.TRACE (49) then -- ignore bad SS releases 
    --          return;
    --       end if;
    --       raise Bad_SS_Mark;

    end Release;
    
    
    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --
 
    procedure Free( Current_TCB  : System.RTS.TCB.Ref_TCB := Null ) is
    	--* OVERVIEW
	--  Free releases the entire secondary stack.  It should be
	--  called at the time a thread is destroyed.

        TCB_Ptr : SYSTEM.RTS.TCB.Ref_TCB := Current_TCB;
	Chunk : Chunk_Ptr;
	Temp_Chunk : Chunk_Ptr;
        use type System.RTS.TCB.Ref_TCB;
    begin
	if TCB_Ptr = Null then
	    TCB_Ptr := SYSTEM.RTS.TCB.Curr_TCB;
	end if;

    	Chunk := To_Chunk_Ptr(TCB_Ptr.Last_Chunk);
	while chunk /= null loop
	    Temp_Chunk := Chunk.Previous;
	    rts_heapfree ( To_Address(Chunk) );
	    Chunk := Temp_Chunk;
	end loop;
    end Free;

    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --


    function Copy_And_Release( From      : in System.Address;
                               Mark_Save : in Mark_Ptr;
                               Size      : in Natural ) return System.Address is
        --* OVERVIEW
        -- Copy_And_Release provides the functionality of "stripping"
        -- the secondary stack back to a given point, similar to Release.
        -- However, it also allows for "except for this data object",
        -- in that a given data object may be made an exception to the
        -- releasing mechanism.
        --
        --* EFFECTS
        -- Copy_And_Release will release all secondary stack back to
        -- Mark_Save, except that space of Size will be reserved on the
	-- secondary stack for the object given at address From.
	-- That object will be copied starting at the point Mark_Save.
	-- Copy_And_Release will return the starting address of the copied 
	-- object.
        -- Copy_And_Release may actually result in an allocation,
        -- as the object to be copied need not necessarily reside in the
        -- secondary stack.  If sufficient space is not available to
        -- accommodate this copy, STORAGE_ERROR will be raised and the
        -- secondary stack state will remain unchanged.
        --
        -- Note that releasing to Mark_Save, allocating Size on the 
	-- secondary stack, and coping the Size object at From will not
        -- be correct if the object to be copied was on the secondary
	-- stack at a point beyond the Mark_Save.
        --
        -- Case 1 : Item at From is on the stack
	--
	-- Before Copy_And_Release
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            | (Previous_Mark)
        --        |        |            |
        --        |        |            | 
        --   direction     |- -- -- -- -| <-- Copy_From (address)
        --      of         | Item       |
        --     stack       |  To        |
        --     growth      |   Copy     |
        --        |        |- -- -- -- -| (Copy_From+Copy_Size)
        --        |        |            |
        --        V        |            |
        --                 |            |
        --                 |------------| Highwater mark (M)
        --                 |            |
        --                       .
        --                       .
        --                       .
	--
	--      Item_Address := Copy_And_Release( 
	--                          From      => Copy_From,
        --                          Mark_Save => Previous_Mark,
        --                          Size      => Copy_Size );
	--
	-- After Copy_And_Release
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |- -- -- -- -| (Previous_Mark) <-- Item_Address
        --        |        |            |
        --        |        |            | 
        --   direction     |            | 
        --      of         |------------| Highwater mark 
        --     stack       |            |
        --     growth      |            |
        --        |        |            |
        --        |        |            |
        --        V        |            |
        --                 |            |
        --                 |            |
        --                 |            |
        --                       .
        --                       .
        --                       .
	--
	-- Copy_And_Release will return the address of the Item in
	-- Item_Address as well as clear all but Item from the stack
	-- and set the new highwater mark.
        --
        --
        -- Case 2 : Item at From is not on the stack
	--
	-- Before Copy_And_Release
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            | (Previous_Mark)
        --        |        |            |
        --        |        |            | 
        --   direction     |------------| Highwater mark (M)
        --      of         |            |
        --     stack       |            |
        --     growth      |            |
        --        |        |            | 
        --        |        |            |
        --        V        |            |
        --                 |            |
        --                 |            | 
        --                 |            |
        --                       .
        --                       .
        --                       .
	--
	-- Somewhere else in memory:
	--
        --                 +------------+ <-- Copy_From
        --                 |            | 
        --                 |            |
        --                 |            | 
        --                 |            |
        --                 |            | 
        --                 |            |
        --                 |            | 
        --                 +------------+ <-- Copy_From+Copy_Size
	--
	--
	--      Item_Address := Copy_And_Release( 
	--                          From      => Copy_From,
        --                          Mark_Save => Previous_Mark,
        --                          Size      => Copy_Size );
	--
	-- After Copy_And_Release
        --                       .
        --                       .
        --                       .
        --                 |            |
        --                 |            | (Previous_Mark) <-- Item_Address
        --        |        |            |
        --        |        |            | 
        --   direction     |            | (M)
        --      of         |            |
        --     stack       |            |
        --     growth      |            |
        --        |        |------------| Highwater mark
        --        |        |            |
        --        V        |            |
        --                 |            |
        --                 |            | 
        --                 |            |
	--
        --                       .
        --                       .
        --                       .
	-- Copy_And_Release will return the address of the Item in
	-- Item_Address as well as clear all but Item from the stack
	-- and set the new highwater mark.  Note that the new highwater
	-- mark is greater than the highwater mark before Copy_And_Release
	-- which means that there was an allocation.

       -- This algorithm has three parts:
       --
       -- 1) Traverse the chain of chunks to get to the chunk that contains
       -- the mark, counting chunks along the way.  This count will be the
       -- used to traverse the chain later when deallocating.
       --
       -- 2) If the object is in the current chunk, then copy it to a
       -- temporary location, set the new highwater mark, and allocate and
       -- copy the temp to the new object.
       --
       -- Otherwise, set the new highwater mark in the chunk, and allocate
       -- and copy the object.  The object will not be clobbered on allocation
       -- so this is ok
       --
       -- 3) Traverse the chain again, deallocating chunks along the way.
       
       Last_Chunk   : Chunk_Ptr := To_Chunk_Ptr (SYSTEM.RTS.TCB.Curr_TCB.Last_Chunk);
       Free_Count   : Natural   := 0; -- number of chunks to free

       -- These are used to restore the SS in case of storage_error;
       Original_Chunk  : Chunk_Ptr      := Last_Chunk;
       Original_HWM    : System.Address := SYSTEM.RTS.TCB.Curr_TCB.Highwater_Mark; 

       Saved_Object    : System.Address;
       New_Object      : System.Address;
       
    begin
       -- this is the case where Initialize failed with Storage_Error
       -- and the Exception_Manager.RTS_SS_Copy_Exception_And_Rlse
       -- calls this routine.  Since SS does not exist just return
       -- From address.
       if Last_Chunk = null then
           return From;
       end if;

       -- Note that if an allocation raises Storage_Error, the state of the
       -- secondary stack must remain unchanged.
       
       -- Traverse the chunk chain until we find the chunk that contains the 
       -- mark, counting the chunks along the way.  
       while Mark_Save.Saved_Sec_Stack_Ptr < To_Address (Last_Chunk) or else
             Mark_Save.Saved_Sec_Stack_Ptr > Last_Chunk.Chunk_End loop
          Last_Chunk := Last_Chunk.Previous;
          Free_Count := Free_Count + 1;

          if Last_Chunk=null then
             raise Bad_SS_Mark;
          end if;
       end loop;


       -- There are two cases: either the object will be clobbered when the 
       -- highwater mark is adjusted for this last chunk, or it won't.  It
       -- will not be clobbered if it is in a chunk the will be freed later
       -- (one traversed above), in the secondary stack at a point "earlier"
       -- than the mark, or is not in the SS at all.  If will be clobbered
       -- if it is in the last chunk at an address greater than the mark.
       
       if (From > To_Address (Last_Chunk) and then
           From < Last_Chunk.Chunk_End) and then
           From >= Mark_Save.Saved_Sec_Stack_Ptr then

          -- The object will be clobbered by the SS release, so make a copy
          Saved_Object := rts_heapalloc (System.Storage_Elements.Storage_Count (Size));
          if Saved_Object = Null_Address then
             raise Storage_Error;
          end if;
          System.Move (To     => Saved_Object, 
                       From   => From, 
                       Length => System.Object_Length (Size));

          SYSTEM.RTS.TCB.Curr_TCB.Highwater_Mark := Mark_Save.Saved_Sec_Stack_Ptr;
          SYSTEM.RTS.TCB.Curr_TCB.Last_Chunk     := To_Address (Last_Chunk);

          -- Allocate and copy the object
          New_Object := Allocate_Without_Exceptions (Size);
          if New_Object = Null_Address then
             -- Restore the SS pointers, free saved object
             SYSTEM.RTS.TCB.Curr_TCB.Last_Chunk     := To_Address (Original_Chunk);
             SYSTEM.RTS.TCB.Curr_TCB.Highwater_Mark := Original_HWM; 
             rts_heapfree (Saved_Object);
             raise Storage_Error;
          end if;
          System.Move (To     => New_Object, 
                       From   => Saved_Object, 
                       Length => System.Object_Length (Size));
          rts_heapfree (Saved_Object);
          
       else
          -- The object will not be clobbered by the adjust, so there's no
          -- need to copy it to a temporarly location first.
          SYSTEM.RTS.TCB.Curr_TCB.Highwater_Mark := Mark_Save.Saved_Sec_Stack_Ptr;
          SYSTEM.RTS.TCB.Curr_TCB.Last_Chunk     := To_Address (Last_Chunk);

          -- Allocate and copy the object
          New_Object := Allocate_Without_Exceptions (Size);
          if New_Object = Null_Address then
             -- Restore the SS pointers          
             SYSTEM.RTS.TCB.Curr_TCB.Last_Chunk     := To_Address (Original_Chunk);
             SYSTEM.RTS.TCB.Curr_TCB.Highwater_Mark := Original_HWM; 
             raise Storage_Error;
          end if;
          System.Move (To     => New_Object, 
                       From   => From, 
                       Length => System.Object_Length (Size));
       end if;

       -- Now that allocation was successful, deallocate all unused chunks
       for i in 1..Free_Count loop
          Last_Chunk := Original_Chunk.Previous;
          rts_heapfree (To_Address (Original_Chunk));
          Original_Chunk := Last_Chunk;
       end loop;
       
       return New_Object;
       
    end Copy_And_Release;
    

    -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * -- * --

end System.RTS.TGT.Sec_Stack_Pkg;
