--   Copyright 1995, Intermetrics, Inc.
--
--
-- Calculator
--   This package implements a calculator which is missing a user interface.
--   To use:
--      -- Create an object of type calculator.calcState.
--      -- Initialize it by calling clearAll.
--      -- Whenever a button is pressed, call buttonPress, passing in
--         the state and the description of the button.
--      -- The generic formal subprogram 'output' will be called whenever
--         a new string needs to be displayed in the calculator's display.
--   

-- This is to suppress to generic elaboration bit,
-- which we don't deal with well at the moment:
pragma Suppress(Elaboration_Check);

with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;

generic
    with procedure output(text : String);
package calculator is

    type ButtonKind is (Digit, Point, Clear, ClearEntry,
			Add, Sub, Mult, Divide, Equals);
    type ButtonOp(kind : ButtonKind) is record
        value : Integer;
    end record;
    type ButtonOpPtr is access all ButtonOp;

    type CalcState is private;

    procedure buttonPress(state : in out CalcState; 
			 button : ButtonOpPtr) ;
    procedure showMsg(state : in out CalcState; msg : String);
    procedure clearAll(state : in out CalcState);

private

    type CalcState is record
	stackedValue : Long_Float;
	stackedOp    : ButtonKind;

	value        : Long_Float;
	image        : Unbounded_String;

	-- display means that image contains a string and value is
	-- meaningless.  Pressing any key will reset the calculator before
	-- the key takes effect.  This is used for startup and for
	-- error messages.
	inDisplay    : Boolean;
	--
	-- not inNumber means that a digit key or point starts a new
	-- number.  inNumber means that a digit key or point appends to
	-- the existing number.
	--
	inNumber     : Boolean;
	pointSeen    : Boolean;

	-- The power of ten value of the last digit pressed, when pointSeen
	-- is true.
	multiplier   : Long_Float;
    end record;

end calculator;


with image; use image;
package body calculator is

    procedure show(state : CalcState);
    procedure clearEntry(state : in out CalcState);
    procedure numberEntry(state : in out CalcState;
			  button : buttonOpPtr);
    function apply(op : ButtonKind; left, right : Long_Float)
		return Long_Float;
    procedure operate(state : in out CalcState);

    -- Refresh the display
    procedure show(state : CalcState) is
    begin
	output(To_String(state.image));
    end;

    -- Display a message, as when an error occurs
    -- As described above under "inDisplay", the next key pressed causes a clear.
    procedure showMsg(state : in out CalcState; msg : String) is
    begin
	state.image := To_Unbounded_String(msg);
	state.inDisplay := true;
	show(state);
    end;

    -- Clear the current entry.
    procedure clearEntry(state : in out CalcState) is
    begin
	state.inDisplay := false;
	state.value := 0.0;
	state.image := To_Unbounded_String("0.");
	state.inNumber := false;
	state.pointSeen := false;
	state.multiplier := 1.0;
    end clearEntry;

    -- Reset the calculator.
    procedure clearAll(state : in out CalcState) is
    begin
	state.stackedValue := 0.0;
	state.stackedOp := Clear;
	clearEntry(state);
    end clearAll;

    -- Handle the input of a digit key or the decimal point.
    procedure numberEntry(state : in out CalcState;
			  button : buttonOpPtr) is
    begin
	if not state.inNumber then
	    clearEntry(state);
	    state.inNumber := true;
	end if;
	if Length(state.image) < 14 then
	    if button.kind = Point then
		state.pointSeen := true;
	    else
		-- Must be a digit
		if state.pointSeen then
		    state.image := state.image & digitimage(button.value);
		    state.multiplier := state.multiplier / 10.0;
		    state.value := state.value +
		      Long_Float(button.value) * state.multiplier;
		else
		    if state.value = 0.0 then
			state.image :=
			  digitimage(button.value) & To_Unbounded_String(".");
		    else
			-- Insert before the decimal point
			Insert(state.image,
			       Before => Length(state.image),
			       New_Item =>
				 String'(1 => digitimage(button.value)));
		    end if;
		    state.value := state.value * 10.0 + Long_Float(button.value);
		end if;
		show(state);
	    end if;
	end if;
    end numberEntry;

    -- Return the result of applying the operator to the two values given
    function apply(op : ButtonKind; left, right : Long_Float)
		return Long_Float is
    begin
	case op is
	    when Add => return left + right;
	    when Sub => return left - right;
	    when Mult => return left * right;
	    when Divide => return left / right;
	    when others => return right;
	end case;
    end apply;

    -- Perform an operation, as when the '=' key is pressed.
    procedure operate(state : in out CalcState) is
    begin
	if state.stackedOp in Add .. Divide then
	    if state.stackedOp = Divide and then state.value = 0.0 then
		showMsg(state, "Error: Divide by 0");
	        return;
	    end if;	
	    state.value :=
	      apply(state.stackedOp, state.stackedValue, state.value);
	    state.image := To_Unbounded_String(double_image(state.value));
	end if;
	state.inNumber := false;
	state.stackedOp := Clear;
	show(state);
    end;

    -- Handle a key press of any kind.
    procedure buttonPress(state : in out CalcState; button : ButtonOpPtr) is
    begin
	if state.inDisplay then
	    clearAll(state);
	end if;
	case button.kind is
	    when Digit | Point =>
		numberEntry(state, button);
	    when Clear =>
		clearAll(state);
		show(state);
	    when ClearEntry =>
		clearEntry(state);
		show(state);
	    when Equals =>
		operate(state);
	    when others =>
		if state.inNumber then
		    operate(state);
		end if;
		state.stackedValue := state.value;
		state.stackedOp := button.kind;
         end case;
    end;
end;
