-- ***************************************************************************
-- *                         Network_IO.ads 
-- * Copyright (c) Simon Johnston & Addison Wesley Longman 1996.
-- *
-- * Description: A specification for a stream oriented network IO package.
-- * Inputs:  None.
-- * Outputs: None.
-- ***************************************************************************

with Ada.Streams;

package Network_IO is

   type Stream_Access is access all
      Ada.Streams.Root_Stream_Type'Class;

   type Connection_Type is private;

   -- This will be extended in time to include
   -- other protocols as support becomes 
   -- available.
   type Connection_Protocol is ( Tcp, Udp );

   -- Create_Connection initialises a server port.
   -- A server port can only be created on this 
   -- host so no host parameter is required.
   function Create_Connection
     (Service	  : in String;
      Protocol	  : in Connection_Protocol;
      Max_Clients  : in Positive := 4)
      return Connection_Type;

   -- This call is used by a server after
   -- calling the Create_Connection above, it 
   -- waits for a connection and returns the client.
   function Wait_For_Client
     (Server_Connection : in Connection_Type)
      return Connection_Type;

   -- Connect_To initialises a client port, and 
   -- as such requires the host on which the server 
   -- port resides.
   function Connect_To
     (Service  : in String;
      Protocol : in Connection_Protocol;
      Host     : in String := "localhost")
      return Connection_Type;

   -- close a connection and all associated streams 
   -- completely.
   procedure Close  
     (Connection       : in out Connection_Type);

   -- This returns a stream type which can be used
   -- with the Read, Write, Input and Output
   -- attributes.
   function Stream(Connection : in Connection_Type) 
      return Stream_Access;

   -- exceptions.
   Connection_Creation_Error : exception;
   Connection_Invalid        : exception;
   Connection_Closed         : exception;
   Operating_System_Error    : exception;
   Parameter_In_Error        : exception;

private
   use  Ada.Streams;

   type Network_Stream_Type is new Root_Stream_Type with
      record
	    Connection : Connection_Type;
      end record;

   type Network_Stream_Access is access all Network_Stream_Type;

   type Connection_Type is new Integer;

   procedure Read
     (Stream : in out Network_Stream_Type;
      Item   : out    Stream_Element_Array;
      Last   : out    Stream_Element_Offset);

   procedure Write
     (Stream : in out Network_Stream_Type;
      Item   : in     Stream_Element_Array);

end Network_IO;
