-- ***************************************************************************
-- *                         Pos.ads
-- * Copyright (c) Simon Johnston & Addison Wesley Longman 1996.
-- *
-- * Description: The Interface to our Point of Sale C API.
-- * Inputs:  None.
-- * Outputs: None.
-- ***************************************************************************

with Interfaces.C;
with Interfaces.C.Pointers;

package Pos is

   type    BYTE        is new Interfaces.C.Unsigned_Char;
   type    PBYTE       is access all BYTE;
   subtype USHORT      is Interfaces.C.Unsigned_Short;
   type    PUSHORT     is access all USHORT;

   type BYTEARRAY1 is array (USHORT range <>) of aliased BYTE;
   package BYTEARRAY2 is 
      new Interfaces.C.Pointers(USHORT,
                                BYTE,
                                BYTEARRAY1,
                                BYTE(0));
   subtype BYTEARRAY is BYTEARRAY2.Pointer;

   package CHARARRAY is 
      new Interfaces.C.Pointers(Interfaces.C.size_t,
                                Interfaces.C.char,
                                Interfaces.C.char_array,
                                Interfaces.C.nul);
   subtype PSZ is CHARARRAY.Pointer;

   -- -------------------------------------------
   -- API Version, used to check calls.
   -- 
   API_VERSION : constant := 255;

   -- -------------------------------------------
   -- Devices.
   --
   type UNIT_ID is (UNIT_NONE,
		    	   UNIT_ALL,
		    	   UNIT_POS_KEYBOARD,
		    	   UNIT_PC_KEYBOARD,
		    	   UNIT_ALARM,
		    	   UNIT_MSR,
		    	   UNIT_KEYLOCK,
		    	   UNIT_MAIN_SCANNER,
		    	   UNIT_HH_SCANNER,
		    	   UNIT_SCALE,
		    	   UNIT_OPER_DISPLAY,
		    	   UNIT_CUST_DISPLAY,
		    	   UNIT_BOTH_DISPLAYS,
		    	   UNIT_RCPT_PRINTER,
		    	   UNIT_AUDT_PRINTER,
		    	   UNIT_BOTH_PRINTERS,
		    	   UNIT_SLIP_PRINTER,
		    	   UNIT_FIRST_CASHDRAWER,
		    	   UNIT_SECOND_CASHDRAWER);

   for UNIT_ID use (UNIT_NONE => -1,
		    	   UNIT_ALL => 0,
		    	   UNIT_POS_KEYBOARD => 1,
		    	   UNIT_PC_KEYBOARD => 2,
		    	   UNIT_ALARM => 3,
		    	   UNIT_MSR => 4,
		    	   UNIT_KEYLOCK => 5,
		    	   UNIT_MAIN_SCANNER => 6,
		    	   UNIT_HH_SCANNER => 7,
		    	   UNIT_SCALE => 8,
		    	   UNIT_OPER_DISPLAY => 9,
		    	   UNIT_CUST_DISPLAY => 10,
		    	   UNIT_BOTH_DISPLAYS => 11,
		    	   UNIT_RCPT_PRINTER => 12,
		    	   UNIT_AUDT_PRINTER => 13,
		    	   UNIT_BOTH_PRINTERS => 14,
		    	   UNIT_SLIP_PRINTER => 15,
		    	   UNIT_FIRST_CASHDRAWER => 16,
		    	   UNIT_SECOND_CASHDRAWER => 17);

   for UNIT_ID'Size use USHORT'Size;

   type PUNIT_ID is access all UNIT_ID;

   -- -------------------------------------------
   -- Error codes.
   --
   type ERROR_CODE is (ERROR_OK,
		       	ERROR_NOT_INITIALISED,
		       	ERROR_ALREADY_INITIALISED,
		       	ERROR_OS_ERROR,
		       	ERROR_COMMAND_INVALID,
		       	ERROR_COMMAND_DATA_ERROR,
		       	ERROR_DEVICE_INVALID,
		       	ERROR_DEVICE_WRITE_FAILED,
		       	ERROR_DEVICE_FAILURE,
		       	ERROR_DEVICE_OFFLINE,
		       	ERROR_DEVICE_SEQUENCE_ERR,
		       	ERROR_DEVICE_STATUS_CHANGED,
		       	ERROR_KEYBOARD_MULTIPLE_KEYS,
		       	ERROR_MSR_INVALID_TRACK,
		       	ERROR_MSR_MULTIPLE_TRACKS,
		       	ERROR_PRINTER_FORMAT_ERR,
	       		ERROR_SCALE_ILLEGAL_WEIGHT,
		       	ERROR_SCALE_ZERO_DETECTED,
		       	ERROR_SCALE_NOT_ZEROED,
		       	ERROR_SCALE_WEIGHT_NOT_REMOVED,
		       	ERROR_SCALE_VERIFY_FAILED,
		       	ERROR_SCALE_WEIGHT_NOT_STABLE,
		       	ERROR_SLIP_SEQUENCE_ERR);

   for ERROR_CODE use (ERROR_OK => 0,
		       	ERROR_NOT_INITIALISED => 1,
		       	ERROR_ALREADY_INITIALISED => 2,
		       	ERROR_OS_ERROR => 3,
		       	ERROR_COMMAND_INVALID => 4,
		       	ERROR_COMMAND_DATA_ERROR => 5,
		       	ERROR_DEVICE_INVALID => 6,
		       	ERROR_DEVICE_WRITE_FAILED => 7,
		       	ERROR_DEVICE_FAILURE => 8,
		       	ERROR_DEVICE_OFFLINE => 9,
		       	ERROR_DEVICE_SEQUENCE_ERR => 10,
		       	ERROR_DEVICE_STATUS_CHANGED => 11,
		       	ERROR_KEYBOARD_MULTIPLE_KEYS => 12,
		       	ERROR_MSR_INVALID_TRACK => 13,
		       	ERROR_MSR_MULTIPLE_TRACKS => 14,
		       	ERROR_PRINTER_FORMAT_ERR => 15,
		       	ERROR_SCALE_ILLEGAL_WEIGHT => 16,
		       	ERROR_SCALE_ZERO_DETECTED => 17,
		       	ERROR_SCALE_NOT_ZEROED => 18,
		       	ERROR_SCALE_WEIGHT_NOT_REMOVED => 19,
		       	ERROR_SCALE_VERIFY_FAILED => 20,
		       	ERROR_SCALE_WEIGHT_NOT_STABLE => 21,
		       	ERROR_SLIP_SEQUENCE_ERR => 22);

   for ERROR_CODE'Size use USHORT'Size;

   type PERROR_CODE is access all ERROR_CODE;

   -- -------------------------------------------
   -- API Control Functions.
   --
   function Initialise
     (ApiVersion :    PUSHORT)
      return ERROR_CODE;

   pragma Import(Stdcall, Initialise, "PosInitialise");

   procedure TerminatePoS
     (TermCode : in USHORT);

   pragma Import(Stdcall, TerminatePoS, "PosTerminate");

   function SystemError
     return ERROR_CODE;

   pragma Import(Stdcall, SystemError, "PosSystemError");

   -- -------------------------------------------
   -- The API Read Function.
   --
   MAX_BUFFER_LENGTH : constant USHORT := 512;

   function Read
     (Buffer	    : in BYTEARRAY;
      BufferLength   : in USHORT;
      ReturnedUnit   : in PUNIT_ID;
      ReturnedLength : in PUSHORT)
      return ERROR_CODE;
   
   pragma Import(Stdcall, Read, "PosRead");
  
   -- -------------------------------------------
   -- General Device Functions.
   --
   type DEVICE_CMD is (RESET,
                       INITIALISE,
                       FLUSH,
                       ONLINE);

   for DEVICE_CMD use (RESET => 0,
                       INITIALISE => 1,
                       FLUSH => 2,
                       ONLINE => 3);

   function DeviceControl
     (Unit    : in  UNIT_ID;
      Command : in DEVICE_CMD)
      return ERROR_CODE;

   pragma Import(Stdcall, DeviceControl, "PosDeviceControl");

   type DEVICE_STATUS is
      record
         DeviceAvailable : Boolean;
         DeviceOnLine    : Boolean;
         HandlerEnabled  : Boolean;
         ExDeviceStatus  : BYTE;
      end record;
   pragma Convention(C, DEVICE_STATUS);

   type PDEVICE_STATUS is access all DEVICE_STATUS;

   for DEVICE_STATUS use
      record
         DeviceAvailable at 0 range 0 .. 1;
         DeviceOnLine    at 0 range 2 .. 3;
         HandlerEnabled  at 0 range 4 .. 5;
         ExDeviceStatus  at 1 range 0 .. 7;
      end record;

   function DeviceStatus
     (Unit    : in UNIT_ID;
      Command : in PDEvICE_STATUS)
      return ERROR_CODE;
 
   pragma Import(Stdcall, DeviceStatus, "PosDeviceStatus");

   -- -------------------------------------------
   -- Alarm Device Functions.
   --
   type ALARM_TONE is
      record
         Duration : USHORT;
         Pitch    : USHORT;
         Volume   : USHORT;
      end record;
   pragma Convention(C, ALARM_TONE);
 
   type ALARM_CONFIG is
      record
         Tone1            : ALARM_TONE;
         Tone2            : ALARM_TONE;
         SupportsTone2     : Boolean;
         SupportsInterrupt : Boolean;
      end record;
   pragma Convention(C, ALARM_CONFIG);
   type PALARM_CONFIG is access all ALARM_CONFIG;

   for ALARM_CONFIG use
      record
         Tone1             at 0  range 0 .. 47;
         Tone2             at 6  range 0 .. 47;
         SupportsTone2     at 12 range 0 .. 1;
         SupportsInterrupt at 12 range 2 .. 3;
      end record;

   function AlarmConfig
     (Unit         : in UNIT_ID;
      ConfigRecord : in PALARM_CONFIG)
      return ERROR_CODE;

   pragma Import(Stdcall, AlarmConfig, "PosAlarmConfig");
 
   function AlarmSound
     (Unit        : in UNIT_ID;
      RepeatValue : in USHORT)
      return ERROR_CODE;

   pragma Import(Stdcall, AlarmSound, "PosAlarmSound");
 
   -- -------------------------------------------
   -- Display Device Functions.
   --
   FADE_ON  : constant := 1;
   FADE_OFF : constant := 0;

   type DISPLAY_CONFIG is
      record
         NumberOfLines     : USHORT;
         CharactersPerLine : USHORT;
         NumberOfLEDs      : USHORT;
         SupportsScrolling : Boolean;
         SupportsFade      : Boolean;
      end record;
   pragma Convention(C, DISPLAY_CONFIG);
   type PDISPLAY_CONFIG is access all DISPLAY_CONFIG;

   for DISPLAY_CONFIG use
      record
         NumberOfLines     at 0 range 0 .. 15;
         CharactersPerLine at 2 range 0 .. 15;
         NumberOfLEDs      at 4 range 0 .. 15;
         SupportsScrolling at 6 range 0 .. 1;
         SupportsFade      at 6 range 2 .. 3;
      end record;

   function DisplayConfig
     (Unit         : in UNIT_ID;
      ConfigRecord : in PDISPLAY_CONFIG)
      return ERROR_CODE;

   pragma Import(Stdcall, DisplayConfig, "PosDisplayConfig");

   function DisplayText 
     (Unit         : in UNIT_ID;
      LineNumber   : in USHORT;
      ColumnNumber : in USHORT;
      TextBuffer   : in PSZ)
      return ERROR_CODE;

   pragma Import(Stdcall, DisplayText, "PosDisplayText");

   function DisplayLed
      (Unit     : in UNIT_ID;
       LedState : in USHORT)
       return ERROR_CODE;

   pragma Import(Stdcall, DisplayLed, "PosDisplayLed");
 
   -- -------------------------------------------
   -- Display Device Functions.
   --
   DRAWER_OPEN   : constant := 1;
   DRAWER_CLOSED : constant := 0;

   function OpenDrawer
     (Unit : in UNIT_ID)
      return ERROR_CODE;

   pragma Import(Stdcall, OpenDrawer, "PosCashdrawerOpen");

   function LockDrawer
     (Unit : in UNIT_ID;
      On   : in Boolean)
      return ERROR_CODE;

   pragma Import(Stdcall, LockDrawer, "PosCashdrawerLock");

   -- -------------------------------------------
   -- MSR not used, binding not included
   --

   -- -------------------------------------------
   -- Port handler not used, binding not included
   --

   -- -------------------------------------------
   -- Display Device Functions.
   --
   PAPER_LOW      : constant := 1;
   PAPER_BACK     : constant := 0;
   SLIP_PAPER_IN  : constant := 2;
   SLIP_PAPER_OUT : constant := 0;

   type PRINTER_CAPS is (EPC_WIDE_ON,
                         EPC_WIDE_OFF,
                         EPC_TALL_ON,
                         EPC_TALL_OFF,
                         EPC_BOLD_ON,
                         EPC_BOLD_OFF,
                         EPC_ITALICS_ON,
                         EPC_ITALICS_OFF,
                         EPC_SIDEWAYS_ON,
                         EPC_SIDEWAYS_OFF,
                         EPC_INVERTED_ON,
                         EPC_INVERTED_OFF);

   for PRINTER_CAPS use (EPC_WIDE_ON => 16,
                         EPC_WIDE_OFF => 17,
                         EPC_TALL_ON => 18,
                         EPC_TALL_OFF => 19,
                         EPC_BOLD_ON => 20,
                         EPC_BOLD_OFF => 21,
                         EPC_ITALICS_ON => 22,
                         EPC_ITALICS_OFF => 23,
                         EPC_SIDEWAYS_ON => 24,
                         EPC_SIDEWAYS_OFF => 25,
                         EPC_INVERTED_ON => 26,
                         EPC_INVERTED_OFF => 27);

   type PRINTER_CONFIG is
      record
         NumberOfLines          : USHORT;
         CharactersPerLine      : USHORT;
         CutMargin              : USHORT;
         SupportsPaperCut       : Boolean;
         SupportsPaperPerforate : Boolean;
         SupportsWide           : Boolean;
         SupportsTall           : Boolean;
         SupportsBold           : Boolean;
         SupportsItalic         : Boolean;
         SupportsSideways       : Boolean;
         SupportsInverted       : Boolean;
      end record;
   pragma Convention(C, PRINTER_CONFIG);
   type PPRINTER_CONFIG is access all PRINTER_CONFIG;

   for PRINTER_CONFIG use
      record
         NumberOfLines          at 0 range 0 .. 15;
         CharactersPerLine      at 2 range 0 .. 15;
         CutMargin              at 4 range 0 .. 15;
         SupportsPaperCut       at 6 range 0 .. 1;
         SupportsPaperPerforate at 6 range 1 .. 2;
         SupportsWide           at 6 range 2 .. 3;
         SupportsTall           at 6 range 3 .. 4;
         SupportsBold           at 6 range 4 .. 5;
         SupportsItalic         at 6 range 5 .. 6;
         SupportsSideways       at 6 range 6 .. 7;
         SupportsInverted       at 6 range 7 .. 8;
      end record;

   function PrinterConfig
     (Unit         : in UNIT_ID;
      ConfigRecord : in PPRINTER_CONFIG)
      return ERROR_CODE;

   pragma Import(Stdcall, PrinterConfig, "PosPrinterConfig");

   function PrintText
     (Unit            : in UNIT_ID;
      LinefeedsBefore : in USHORT;
      LinefeedsAfter  : in USHORT;
      TextBuffer      : in PSZ)
      return ERROR_CODE;

   pragma Import(Stdcall, PrintText, "PosPrintText");

   function PrinterAdvance
     (Unit          : in UNIT_ID;
      NumberOfLines : in BYTE)
      return ERROR_CODE;

   pragma Import(Stdcall, PrinterAdvance, "PosPrinterAdvance");

   function PrinterCut
     (Unit          : in UNIT_ID;
      PerforateOnly : in Boolean)
      return ERROR_CODE;

   pragma Import(Stdcall, PrinterCut, "PosPrinterCut");

   type SLIPPRINTER_CMD is (OPEN_JAWS,
                            CLOSE_JAWS,
                            ROLL_PAPER_IN,
                            EJECT_PAPER);

   for SLIPPRINTER_CMD use (OPEN_JAWS => 0,
                            CLOSE_JAWS => 1,
                            ROLL_PAPER_IN => 2,
                            EJECT_PAPER => 3);

   function SlipPrinterCommand
     (Unit    : in UNIT_ID;
      Command : in SLIPPRINTER_CMD)
      return ERROR_CODE;

   pragma Import(Stdcall, SlipPrinterCommand,
                          "PosSlipPrinterCommand");

   -- -------------------------------------------
   -- Weigh Scale not used, binding not included
   --

end Pos;
