--[class_histogram.ads] Specification
WITH Simple_io, Ada.Characters.Handling; 
USE  Simple_io, Ada.Characters.Handling;
PACKAGE Class_histogram IS
  TYPE Histogram IS PRIVATE;
  DEF_HEIGHT : CONSTANT Positive := 14;

  PROCEDURE reset( the:IN OUT Histogram );
  PROCEDURE add_to( the:IN OUT Histogram; a_ch:IN Character );
  PROCEDURE put(the:IN Histogram; height:IN Positive:=DEF_HEIGHT);
PRIVATE
  TYPE    Alphabet_index IS NEW Character RANGE 'A' .. 'Z';
  SUBTYPE Alphabet_range IS Alphabet_index;
  TYPE    Alphabet_array IS ARRAY (Alphabet_range) OF Natural;

  TYPE Histogram IS RECORD
    number_of   : Alphabet_array := ( OTHERS => 0 );
  END RECORD;
END Class_histogram;

--[class_histogram.adb] Implementation
PACKAGE BODY Class_histogram IS

  PROCEDURE reset(the:IN OUT Histogram) IS
  BEGIN
    the.number_of := ( OTHERS => 0 );  -- Reset counts to 0
  END reset;

  PROCEDURE add_to(the:IN OUT Histogram; a_ch:IN Character) IS
    ch : Character;
  BEGIN
    ch := a_ch;                        -- As write to ch
    IF is_lower(ch) THEN               -- Convert to upper case
      ch := to_upper( ch );
    END IF;
    IF is_upper( ch ) THEN             -- so record
      DECLARE
        c : Alphabet_index := Alphabet_index(ch);
      BEGIN
        the.number_of(c) := the.number_of(c) + 1;
      END;
    END IF;
  END add_to;

  PROCEDURE put(the:IN Histogram; 
                height:IN Positive:=DEF_HEIGHT) IS
    frequency    : Alphabet_array;        -- Copy to process
    max_height   : Natural := 0;          -- Observed max
  BEGIN
    frequency := the.number_of;           -- Copy data (Array)
    FOR ch IN Alphabet_range LOOP         -- Find max frequency
      IF frequency(ch) > max_height THEN
        max_height:= frequency(ch);
      END IF;
    END LOOP;

    IF max_height > 0 THEN
      FOR ch IN Alphabet_range LOOP      -- Scale to max height
        frequency(ch):=(frequency(ch)*height)/(max_height);
      END LOOP;
    END IF;

    FOR row IN REVERSE 1 .. height LOOP  -- Each line
      put( "  | " );                     -- start of line
      FOR ch IN Alphabet_range LOOP
        IF frequency(ch) >= row THEN
          put('*');                      -- bar of hist >= col
        ELSE
          put(' ');                      -- bar of hist <  col
        END IF;
      END LOOP;
      put(" | "); new_line;              -- end of line
    END LOOP;
    put("  +----------------------------+"); new_line;
    put("    ABCDEFGHIJKLMNOPQRSTUVWXYZ " ); new_line;
    put("  *  = (approx) ");
    put( Float(max_height) / Float(height), aft=>2, exp=>0 );
    put(" characters "); new_line;
  END put;
END Class_histogram;

--[main.adb] Procedure
WITH Simple_io, Class_histogram;
USE  Simple_io, Class_histogram;
PROCEDURE main080 IS
  ch:Character;                     -- Current character
  text_histogram: Histogram;        -- Histogram object
BEGIN
  reset(text_histogram);            -- Reset to empty

  WHILE NOT end_of_file LOOP        -- For each line
    WHILE NOT end_of_line LOOP      -- For each character
      get(ch);                      -- Get current character
      add_to( text_histogram, ch ); -- Add to histogram
    END LOOP;
    skip_line;                      -- Next line
  END LOOP;

  put( text_histogram );            -- Print histogram

END main080;
