--[pack_types.ads] Specification
WITH Ada.Text_io;
USE  Ada.Text_io;
PACKAGE Pack_types190 IS
  TYPE P_File_type IS ACCESS ALL Ada.Text_io.File_type;
  EOT  : CONSTANT Character := Character'Val(0);
  CR   : CONSTANT Character := Character'Val(15);
  QUEUE_SIZE    : CONSTANT := 3;

  TYPE    Queue_no    IS NEW Integer   RANGE 0 .. QUEUE_SIZE;
  TYPE    Queue_index IS MOD QUEUE_SIZE;
  SUBTYPE Queue_range IS Queue_index;
  TYPE    Queue_array IS ARRAY ( Queue_range ) OF Character;
END Pack_types190;

--[pack_threads.ads] Specification
WITH Pack_types190;
USE  Pack_types190;
PACKAGE Pack_threads190 IS
  PROTECTED TYPE PT_buffer IS         -- Task type specification
    ENTRY put( ch:IN Character; no_more:IN Boolean );
    ENTRY get( ch:IN OUT Character; no_more:OUT Boolean);
  PRIVATE
    elements    : Queue_array;              -- Array of elements
    head        : Queue_index := 0;         -- Index
    tail        : Queue_index := 0;         -- Index
    no_in_queue : Queue_no    := 0;         -- Number in queue
    fin         : Boolean     := FALSE;     -- Finish;
  END PT_buffer;

  TYPE P_PT_buffer IS ACCESS ALL PT_buffer;

  TASK TYPE Task_read(p_buffer:P_PT_buffer; fd_in:P_File_type) IS
    ENTRY finish;
  END Task_read;

  TASK TYPE Task_write(p_buffer:P_PT_buffer; fd_out:P_File_type) IS
    ENTRY finish;
  END Task_write;
END Pack_threads190;

--[pack_threads.adb] Implementation
WITH Ada.Text_io;
USE  Ada.Text_io;
PACKAGE BODY Pack_threads190 IS

  PROTECTED BODY PT_buffer IS

    ENTRY put( ch:IN Character; no_more:IN Boolean )
      WHEN no_in_queue < QUEUE_SIZE  IS
    BEGIN
        IF no_more THEN                    -- Last
          fin := TRUE;                     -- Set flag
        ELSE
          elements( tail ) := ch;          -- Add to queue
          tail := tail+1;                  -- Next position
          no_in_queue := no_in_queue + 1;  --
        END IF;
    END;

    ENTRY get(ch:IN OUT Character; no_more:OUT Boolean)
      WHEN no_in_queue > 0 OR ELSE fin IS
    BEGIN
      IF no_in_queue > 0 THEN             -- Item available
        ch := elements( head );           -- Get item
        head := head+1;                   -- Next position
        no_in_queue := no_in_queue - 1;   --
        no_more := FALSE;                 -- Not end
      ELSE
        no_more := TRUE;                  -- End of items
      END IF;
    END;

  END PT_buffer;

  TASK BODY Task_read IS                  -- Task implementation
    ch      : Character;
  BEGIN
    WHILE NOT end_of_file( fd_in.ALL ) LOOP
      WHILE NOT end_of_line( fd_in.ALL ) LOOP
        get( fd_in.ALL, ch);              -- Get character
        p_buffer.put( ch, FALSE );        -- Add to buffer
        --DELAY 0.00001;                  -- Cause task switch
      END LOOP;
      skip_line( fd_in.ALL );             -- Next line
      p_buffer.put( CR, FALSE );          -- New line
      --DELAY 0.00001;                    -- Cause task switch
    END LOOP;
    p_buffer.put( EOT, TRUE );            -- End of characters

    ACCEPT finish;
  EXCEPTION
    WHEN Tasking_error =>
      put("Exception in Task read"); new_line;
  END Task_read;

  TASK BODY Task_write IS                 -- Task implementation
    last     : Boolean := FALSE;          -- No more data
    ch       : Character;                 -- Character read
  BEGIN
    LOOP
      p_buffer.get( ch, last );           -- From buffer
      EXIT WHEN last;                     -- No more characters
      IF ch = CR THEN
        new_line( fd_out.ALL );           -- New line
      ELSE
        put( fd_out.ALL, ch );            -- Character
      END IF;
      --DELAY 0.00001;                    -- Cause task switch
    END LOOP;

    ACCEPT finish;                        -- Finished
  EXCEPTION
    WHEN Tasking_error =>
      put("Exception in Task write"); new_line;
  END Task_write;
END Pack_threads190;

--[copy.adb] Procedure
WITH Ada.Text_io, Pack_threads190, Pack_types190;
USE  Ada.Text_io, Pack_threads190, Pack_types190;
PROCEDURE do_copy(from:IN String; to:IN String) IS
  TYPE State IS ( OPEN_FILE, CREATE_FILE );
  fd_in   : P_File_Type := NEW Ada.Text_io.File_type;
  fd_out  : P_File_type := NEW Ada.Text_io.File_type;
  mode    : State := OPEN_FILE;
BEGIN
  open(  File=>fd_in.ALL,  Mode=>IN_FILE,  Name=>from);
  mode := CREATE_FILE;
  create(File=>fd_out.ALL, Mode=>OUT_FILE, Name=>to);
  DECLARE
    buffers : P_PT_buffer := NEW PT_buffer;
    reader  : Task_read( buffers, fd_in );
    writer  : Task_write( buffers, fd_out );
  BEGIN
    reader.finish;  close( fd_in.ALL );   -- Finish reader task
    writer.finish;  close( fd_out.ALL );  -- Finish writer task
  END;
EXCEPTION
  WHEN Name_error =>
    CASE mode IS
      WHEN OPEN_FILE =>
        put("Problem opening file " & from ); new_line;
      WHEN CREATE_FILE =>
        put("Problem creating file " & to ); new_line;
    END CASE;
  WHEN Tasking_error =>
    put("Task error in main program"); new_line;
END do_copy;

--[main.adb] Procedure
WITH do_copy;
WITH Ada.Text_io, Ada.Command_line;
USE  Ada.Text_io, Ada.Command_line;
PROCEDURE copy190 IS
BEGIN
  IF argument_count = 2 THEN
    do_copy ( argument(1), argument(2) );
  ELSE
    put("Usage: copy from to"); new_line;
  END IF;
END copy190;
