 --[pack_screen.ads] Specification                   -- End of line
PACKAGE Pack_screen IS
  PROCEDURE screen_clear;                -- Home clear screen
  PROCEDURE screen_home;                 -- Home no clear screen
PRIVATE
  ESC: CONSTANT Character := Character'Val(27);
END Pack_screen;

--[pack_screen.adb] Implementation
WITH Text_io; USE Text_io;
PACKAGE BODY Pack_screen IS             -- Terminal dependent I/O
  PROCEDURE screen_clear IS             -- Clear screen
  BEGIN
    put( ESC & "[2J" );                 -- Escape sequence
  END screen_clear;
  PROCEDURE screen_home IS              -- Home
  BEGIN
    put( ESC & "[0;0H");                -- Escape sequence
  END screen_home;
END Pack_screen;

----------------------------------------------------------------------

--[class_counter.ads] Specification
PACKAGE Class_counter IS
  TYPE Counter        IS PRIVATE;
  TYPE Counter_colour IS ( BLACK, WHITE );
  PROCEDURE set( the:IN OUT Counter; rep:IN Counter_colour );
  PROCEDURE display( the:IN Counter );
  PROCEDURE display_none( the:IN Counter );
  PROCEDURE flip( the:IN OUT Counter );
  FUNCTION  rep( the:IN Counter ) RETURN Counter_colour;
PRIVATE
  TYPE Counter IS RECORD
    colour: Counter_colour;         -- Colour of counter
  END RECORD;
END Class_counter;

--[class_counter.adb] Implementation
WITH Simple_io; USE Simple_io;
PACKAGE BODY Class_counter IS
  PROCEDURE set( the:IN OUT Counter; rep:IN Counter_colour ) IS
  BEGIN
    the.colour := rep;
  END set;

  PROCEDURE display( the:IN Counter ) IS
  BEGIN
    CASE the.colour IS
      WHEN BLACK  => put('X'); -- Representation of a black piece
      WHEN WHITE  => put('O'); -- Representation of a white piece
    END CASE;
  END display;

  PROCEDURE display_none( the:IN Counter ) IS
  BEGIN
    put(' ');                   -- Representation of NO piece
  END display_none;

  PROCEDURE flip( the:IN OUT Counter ) IS
  BEGIN
    CASE the.colour IS
      WHEN BLACK  => the.colour := WHITE;     -- Flip to White
      WHEN WHITE  => the.colour := BLACK;     -- Flip to Black
    END CASE;
  END flip;

  FUNCTION  rep( the:IN Counter ) RETURN Counter_colour IS
  BEGIN
    RETURN the.colour;  -- Representation of the counter colour
  END rep;
END Class_counter;

----------------------------------------------------------------------

--[class_cell.ads] Specification
WITH Class_counter;
USE  Class_counter;
PACKAGE Class_cell IS
  TYPE Cell IS PRIVATE;
  TYPE Cell_holds IS ( C_WHITE, C_BLACK, EMPTY );

  PROCEDURE initialize( the:IN OUT Cell );
  FUNCTION  holds( the:IN Cell ) RETURN Cell_holds;
  PROCEDURE add( the:IN OUT Cell; players_counter:IN Counter );
  PROCEDURE display( the:IN Cell );
  PROCEDURE flip( the:IN OUT Cell );
  FUNCTION to_colour( c:IN Cell_holds ) RETURN Counter_colour;
PRIVATE
  TYPE Cell_is IS ( EMPTY_CELL, NOT_EMPTY_CELL );
  TYPE Cell IS RECORD
    contents: Cell_is := EMPTY_CELL;
    item    : Counter;                 -- The counter
  END RECORD;
END Class_cell;

--[class_cell.adb] Implementation
PACKAGE BODY Class_cell IS
  PROCEDURE initialize( the:IN OUT Cell ) IS
  BEGIN
    the.contents := EMPTY_CELL;   -- Initialize cell to empty
  END initialize;

  FUNCTION  holds( the:IN Cell ) RETURN Cell_holds IS
  BEGIN
    CASE the.contents IS
      WHEN EMPTY_CELL     =>             -- Empty
        RETURN EMPTY;                    --  No counter
      WHEN NOT_EMPTY_CELL =>             -- Counter
        CASE rep( the.item ) IS
          WHEN WHITE => RETURN C_WHITE;  --  white counter
          WHEN BLACK => RETURN C_BLACK;  --  black counter
        END CASE;
    END CASE;
  END holds;

  PROCEDURE add(the:IN OUT Cell; players_counter:IN Counter) IS
  BEGIN
    the := (NOT_EMPTY_CELL,players_counter);
  END add;

  PROCEDURE display( the:IN Cell ) IS
  BEGIN
    IF the.contents = NOT_EMPTY_CELL THEN
      display( the.item );           -- Display the counter
    ELSE
      display_none( the.item );      -- No counter
    END IF;
  END display;

  PROCEDURE flip( the:IN OUT Cell ) IS
  BEGIN
    flip( the.item );                -- Flip counter
  END flip;

  FUNCTION to_colour(c:IN Cell_holds) RETURN Counter_colour IS
  BEGIN
    CASE c IS                        -- Conversion of enum.
      WHEN C_WHITE => RETURN WHITE;
      WHEN C_BLACK => RETURN BLACK;
      WHEN OTHERS  => RAISE Constraint_error;
    END CASE;
  END to_colour;

END Class_cell;

----------------------------------------------------------------------

--[class_board.ads] Specification
WITH Class_counter, Class_cell;
USE  Class_counter, Class_cell;
PACKAGE Class_board310 IS

  TYPE Board         IS PRIVATE;
  TYPE State_of_game IS ( PLAY, WIN, DRAW, LOSE );
  TYPE Move_status   IS ( OK, INVALID, PASS );

  PROCEDURE set_up( the:IN OUT Board );
  PROCEDURE add( the:IN OUT Board; x,y:IN Integer;
                 move_is:IN Move_status );
  PROCEDURE now_playing( the:IN OUT Board; c:IN Counter_colour );
  PROCEDURE display( the:IN Board );
  FUNCTION  check_move( the:IN Board; x,y:IN Integer )
                        RETURN Move_status;
  FUNCTION  status( the:IN Board ) RETURN State_of_game;
  FUNCTION  contents( the:IN Board; x,y:IN Integer )
                     RETURN Cell_holds;

PRIVATE
  SIZE: CONSTANT := 8;                            -- 8 * 8 Board
  SUBTYPE Board_index IS Integer RANGE 1 .. SIZE; --
  SUBTYPE Board_range IS Board_index;             --

  TYPE Board_array IS ARRAY (Board_range, Board_range) OF Cell;
  TYPE Score_array IS ARRAY (Counter_colour) OF Natural;
  TYPE Move_array  IS ARRAY (Counter_colour) OF Move_status;

  TYPE Board IS RECORD
    sqrs     : Board_array;               -- Game board
    player   : Counter_colour;            -- Current Player
    opponent : Counter_colour;            -- Opponent
    score    : Score_array;               -- Running score
    last_move: Move_array;                -- Last move is
  END RECORD;
END Class_board310;

--[class_board.adb] Implementation
WITH Simple_io, Pack_screen;
USE  Simple_io, Pack_screen;
PACKAGE BODY Class_board310 IS

  PROCEDURE next( the:IN Board; x_co,y_co:IN OUT Board_index;
                  dir:IN Natural; res:OUT Boolean);
  FUNCTION find_turned( the:IN Board; x,y: IN Board_index )
                        RETURN Natural;
  PROCEDURE turn_counters(the: IN OUT Board; x,y: IN Board_index;
                          total: OUT Natural );
  FUNCTION  no_turned(the:IN Board; o_x,o_y:IN Board_index;
                      dir:IN Natural;
                      n:IN Natural := 0 ) RETURN Natural;
  PROCEDURE capture(the:IN OUT Board; x_co, y_co:IN Board_index;
                    dir:IN Natural );

  PROCEDURE set_up( the:IN OUT Board ) IS
    black_counter: Counter;               -- A white counter
    white_counter: Counter;               -- A black counter
  BEGIN
    set( black_counter, BLACK );          -- Set black
    set( white_counter, WHITE );          -- Set white
    FOR x IN the.sqrs'Range(1) LOOP
      FOR y IN the.sqrs'Range(2) LOOP
        initialize( the.sqrs(x,y) );      -- To empty
      END LOOP;
    END LOOP;
    add( the.sqrs( SIZE/2,   SIZE/2 ),   black_counter );
    add( the.sqrs( SIZE/2,   SIZE/2+1 ), white_counter );
    add( the.sqrs( SIZE/2+1, SIZE/2 ),   white_counter );
    add( the.sqrs( SIZE/2+1, SIZE/2+1 ), black_counter );
    the.score( BLACK ) := 2; the.score( WHITE ) := 2;
  END set_up;

  PROCEDURE add( the:IN OUT Board; x,y:IN Integer;
                 move_is:IN Move_status ) IS
    plays_with: Counter;          -- Current player's counter
    turned    : Natural;          -- Number counters turned
  BEGIN
    set( plays_with, the.player );          -- Set current colour
    the.last_move( the.player ) := move_is; -- Last move is
    IF move_is = OK THEN                    -- Not Pass
      turn_counters(the, x,y, turned);      -- and flip
      add( the.sqrs( x, y ), plays_with );  -- to board
      the.score( the.player ) :=
                 the.score( the.player ) + turned + 1;
      the.score( the.opponent ):=
                 the.score( the.opponent ) - turned;
    END IF;
  END add;

  PROCEDURE now_playing(the:IN OUT Board; c:IN Counter_colour) IS
  BEGIN
    the.player   := c;                       -- Player
    CASE c IS                                -- Opponent
       WHEN WHITE => the.opponent := BLACK;
       WHEN BLACK => the.opponent := WHITE;
    END CASE;
  END now_playing;

  PROCEDURE display( the:IN Board ) IS
    dashes: String( 1 .. the.sqrs'Length*4+1 ) := (OTHERS=>'-');
  BEGIN
    screen_clear;                            -- Clear screen
    put( dashes ); new_line;                 -- Top
    FOR x IN the.sqrs'Range(1) LOOP
      put("|");                              -- Cells on line
      FOR y IN the.sqrs'Range(2) LOOP
        put(" "); display( the.sqrs(x,y) ); put(" |");
      END LOOP;
      new_line; put( dashes ); new_line;     -- Bottom lines
    END LOOP;
    new_line;
    put( "Player X has " );
    put( Integer(the.score(BLACK)), width=>2 );
    put( " counters" ); new_line;
    put( "Player O has " );
    put( Integer(the.score(WHITE)), width=>2 );
    put( " counters" ); new_line;
  END display;

  FUNCTION check_move( the:IN Board; x,y:IN Integer )
                       RETURN Move_status IS
  BEGIN
    IF x = 0 AND THEN y = 0 THEN
      RETURN PASS;
    ELSIF x IN Board_index AND THEN y IN Board_index THEN
      IF holds( the.sqrs( x, y ) ) = EMPTY THEN
        IF find_turned(the, x, y) > 0 THEN
          RETURN OK; 
        END IF;
      END IF;
    END IF;
    RETURN INVALID;
  END check_move;

  FUNCTION find_turned( the:IN Board; x,y: IN Board_index )
                        RETURN Natural IS
    sum     : Natural := 0;       -- Total stones turned
  BEGIN
    IF holds( the.sqrs( x, y ) ) = EMPTY THEN
      FOR dir IN 1 .. 8 LOOP      -- The  8 possible directions
        sum := sum + no_turned( the, x, y, dir );
      END LOOP;
    END IF;
    RETURN sum;                   -- return total
  END find_turned;

  PROCEDURE turn_counters(the: IN OUT Board; x,y: IN Board_index;
                          total: OUT Natural ) IS
    num_cap  : Natural := 0;
    captured : Natural;
  BEGIN
    IF holds( the.sqrs( x, y ) ) = EMPTY THEN
      FOR dir IN 1 .. 8 LOOP
        captured := no_turned( the, x, y, dir );
        IF captured > 0 THEN
          capture( the, x, y, dir );
          num_cap := num_cap + captured;
        END IF;
      END LOOP;
    END IF;
    total := num_cap;
  END turn_counters;

  FUNCTION  no_turned(the:IN Board; o_x,o_y:IN Board_index;
                      dir:IN Natural;
                      n:IN Natural := 0 ) RETURN Natural IS
    ok : Boolean;                           -- Result from next
    nxt: Cell_holds;                        -- Next in line is
    col: Counter_colour;                    -- Counter colour
    x  : Board_index := o_x;                -- Local copy
    y  : Board_index := o_y;                -- Local copy
  BEGIN
    next( the, x,y, dir, ok );              -- Next cell
    IF ok THEN                              -- On the board
      nxt := holds( the.sqrs(x,y) );        -- Contents are
      IF nxt = EMPTY THEN                   -- End of line
        RETURN 0;
      ELSE
        col := to_colour( nxt );            -- Colour
        IF col = the.opponent THEN          -- Opponents counter
          RETURN no_turned(the, x,y, dir, n+1); -- Try next cell
        ELSIF col = the.player THEN         -- End of counters
          RETURN n;                         -- Counters turned
        END IF;
      END IF;
    ELSE
      RETURN 0;                             -- No line
    END IF;
  END no_turned;

  PROCEDURE next( the:IN Board; x_co,y_co:IN OUT Board_index;
                  dir:IN Natural; res:OUT Boolean) IS
    x, y   : Natural;
  BEGIN
    x := x_co; y := y_co;         -- May go outside Board_range
    CASE dir IS
      WHEN 1 =>         y:=y+1;   --  Direction to move
      WHEN 2 => x:=x+1; y:=y+1;   --      8   1   2
      WHEN 3 => x:=x+1;           --
      WHEN 4 => x:=x+1; y:=y-1;   --      7   *   3
      WHEN 5 =>         y:=y-1;   --
      WHEN 6 => x:=x-1; y:=y-1;   --      6   5   4
      WHEN 7 => x:=x-1;           --
      WHEN 8 => x:=x-1; y:=y+1;   --
      WHEN OTHERS => RAISE Constraint_error;
    END CASE;
    IF x IN Board_range AND THEN y IN Board_range THEN
      x_co :=  x; y_co :=  y;     --
      res := TRUE;                -- Found a next cell
    ELSE
      res := FALSE;               -- No next cell
    END IF;
  END next;

  PROCEDURE capture(the:IN OUT Board; x_co, y_co:IN Board_index;
             dir:IN Natural ) IS
    ok   : Boolean;                   -- There is a next cell
    x, y : Board_index;               -- Coordinates of cell
    nxt  : Cell_holds;                -- Next in line is
  BEGIN
    x := x_co; y := y_co;
    next( the, x, y, dir, ok );       -- Calculate pos next cell
    IF ok THEN                        -- Cell exists (Must)
      nxt := holds( the.sqrs(x,y) );
      IF to_colour( nxt ) = the.opponent THEN
        flip( the.sqrs(x, y) );       -- Capture
        capture(the, x, y, dir );     -- Implement capture
      ELSE
        RETURN;                       -- End of line
      END IF;
    ELSE
      RAISE Constraint_error;         -- Will never occur
    END IF;
  END capture;

  FUNCTION  status ( the:IN Board ) RETURN State_of_game IS
  BEGIN
    IF the.score( the.opponent ) = 0 THEN
      RETURN WIN;
    END IF;
    IF (the.sqrs'Length(1) * the.sqrs'Length(2) =
        the.score(the.opponent)+the.score(the.player)) OR
       (the.last_move(BLACK)=PASS AND the.last_move(WHITE)=PASS)
    THEN
      IF the.score(the.opponent) = the.score(the.player)
         THEN RETURN DRAW;
      END IF;
      IF the.score(the.opponent) < the.score(the.player)
         THEN RETURN WIN;
      ELSE
        RETURN LOSE;
      END IF;
    END IF;
    RETURN PLAY;
  END;

  FUNCTION contents( the:IN Board; x,y:IN Integer )
                     RETURN Cell_holds IS
  BEGIN
      RETURN holds( the.sqrs( x, y ) );
  END contents;

END Class_board310;

----------------------------------------------------------------------

--[class_player.ads] Specification
WITH Class_counter, Class_board310;
USE  Class_counter, Class_board310;
PACKAGE Class_player IS
  TYPE Player IS PRIVATE;

  PROCEDURE set( the:IN OUT Player; c:IN Counter_colour );
  PROCEDURE get_move(the:IN Player; row,column:OUT Integer);
  FUNCTION  my_counter( the:IN Player ) RETURN Counter;
  PROCEDURE announce( the:IN Player; what:IN State_of_game );
PRIVATE
  TYPE Player IS RECORD
    plays_with : Counter;        -- Player's counter
  END RECORD;
END Class_player;

--[class_player.adb] Implementation
WITH Simple_io;
USE  Simple_io;
PACKAGE BODY Class_player IS
  PROCEDURE set(the:IN OUT Player; c:IN Counter_colour ) IS
    a_counter : Counter;
  BEGIN
    set( a_counter, c );          -- Set colour
    the.plays_with := a_counter;  -- Player is playing with
  END set;

  PROCEDURE get_move(the:IN Player; row,column:OUT Integer) IS
    valid_move : Boolean := FALSE;
  BEGIN
    WHILE NOT valid_move LOOP
      BEGIN
        put("Please enter move "); display( the.plays_with );
        put(" row column : "); get( row ); get( column );
        valid_move := TRUE;
      EXCEPTION
        WHEN Data_Error =>
          row := -1; column := -1; skip_line;
        WHEN End_error =>
          row := 0; column := 0;
          RETURN;
      END;
    END LOOP;
  END get_move;

  FUNCTION  my_counter( the:IN Player ) RETURN Counter IS
  BEGIN
    RETURN the.plays_with;
  END my_counter;

  PROCEDURE announce(the:IN Player; what:IN State_of_game) IS
  BEGIN
    CASE what IS
      WHEN WIN    =>
        put("Player "); display( the.plays_with );
        put(" has won");
      WHEN LOSE   =>
        put("Player "); display( the.plays_with );
        put(" has lost");
      WHEN DRAW   =>
        put("It's a draw");
      WHEN OTHERS =>
        RAISE Constraint_error;
    END CASE;
    new_line;
  END announce;

END Class_player;

----------------------------------------------------------------------

--[play.adb] Procedure
WITH Class_board310, Class_player, Class_counter;
USE  Class_board310, Class_player, Class_counter;
PROCEDURE play IS                            -- Reversi
  reversi       : Board;                     -- The playing board
  contestant    : ARRAY(Counter_colour) OF Player;
  current_state : State_of_game;             -- State of game
  person        : Counter_colour;            -- Current player
  x, y          : Integer;                   -- Move
  move_is       : Move_status;               -- Last move is
BEGIN
  set_up( reversi );                         -- Set up board
  set( contestant(BLACK), BLACK );           -- Set player black
  set( contestant(WHITE), WHITE );           -- Set player white

  current_state := PLAY;  person := BLACK;   -- Black starts

  display( reversi );                        -- Initial board

  WHILE current_state = PLAY LOOP            -- Playable game
    now_playing( reversi, person );          -- set player

    LOOP                                     -- Get move
      get_move( contestant(person), x, y );
      move_is := check_move(reversi, x, y);  -- Validate
      EXIT WHEN move_is=OK OR move_is=PASS;  -- OK
    END LOOP;

    add( reversi, x, y, move_is );           -- Add move to board

    display( reversi );                      -- Display new board
    current_state := status( reversi );      -- State of play is

    IF current_state = PLAY THEN             -- Is still playable
      CASE person IS                         -- next player
        WHEN BLACK  => person := WHITE;
        WHEN WHITE  => person := BLACK;
      END CASE;
    END IF;

  END LOOP;                                   -- Next move

  announce( contestant(person), current_state );  -- Result

END play;
